# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides support for importing and exporting SVG files.

It enables importing/exporting objects directly to/from the 3D document
but doesn't handle the SVG output from the TechDraw module.

Currently it only reads the following entities:
* paths, lines, circular arcs, rects, circles, ellipses, polygons, polylines.

Currently unsupported:
* use, image.
"""
## @package importSVG
#  \ingroup DRAFT
#  \brief SVG file importer and exporter

# Check code with
# flake8 --ignore=E226,E266,E401,W503

__title__ = "FreeCAD Draft Workbench - SVG importer/exporter"
__author__ = "Yorik van Havre, Sebastian Hoogen"
__url__ = "https://www.freecad.org"

# TODO:
# ignoring CDATA
# handle image element (external references and inline base64)
# debug Problem with 'Sans' font from Inkscape
# debug Problem with fill color
# implement inheriting fill style from group
# handle relative units

import math
import os
import re
import xml.sax

import FreeCAD
import Part
import Draft
from DraftVecUtils import equals
from FreeCAD import Vector
from draftutils import params
from draftutils import utils
from draftutils.utils import svg_precision
from draftutils.translate import translate
from draftutils.messages import _err, _msg, _wrn
from draftutils.utils import pyopen
from SVGPath import SvgPathParser
import xml.etree.ElementTree as ET
from copy import deepcopy

if FreeCAD.GuiUp:
    from PySide import QtWidgets
    import FreeCADGui

    gui = True
    try:
        draftui = FreeCADGui.draftToolBar
    except AttributeError:
        draftui = None
else:
    gui = False
    draftui = None


svgcolors = {
    "Pink": (255, 192, 203),
    "Blue": (0, 0, 255),
    "Honeydew": (240, 255, 240),
    "Purple": (128, 0, 128),
    "Fuchsia": (255, 0, 255),
    "LawnGreen": (124, 252, 0),
    "Amethyst": (153, 102, 204),
    "Crimson": (220, 20, 60),
    "White": (255, 255, 255),
    "NavajoWhite": (255, 222, 173),
    "Cornsilk": (255, 248, 220),
    "Bisque": (255, 228, 196),
    "PaleGreen": (152, 251, 152),
    "Brown": (165, 42, 42),
    "DarkTurquoise": (0, 206, 209),
    "DarkGreen": (0, 100, 0),
    "MediumOrchid": (186, 85, 211),
    "Chocolate": (210, 105, 30),
    "PapayaWhip": (255, 239, 213),
    "Olive": (128, 128, 0),
    "Silver": (192, 192, 192),
    "PeachPuff": (255, 218, 185),
    "Plum": (221, 160, 221),
    "DarkGoldenrod": (184, 134, 11),
    "SlateGrey": (112, 128, 144),
    "MintCream": (245, 255, 250),
    "CornflowerBlue": (100, 149, 237),
    "Gold": (255, 215, 0),
    "HotPink": (255, 105, 180),
    "DarkBlue": (0, 0, 139),
    "LimeGreen": (50, 205, 50),
    "DeepSkyBlue": (0, 191, 255),
    "DarkKhaki": (189, 183, 107),
    "LightGrey": (211, 211, 211),
    "Yellow": (255, 255, 0),
    "Gainsboro": (220, 220, 220),
    "MistyRose": (255, 228, 225),
    "SandyBrown": (244, 164, 96),
    "DeepPink": (255, 20, 147),
    "Magenta": (255, 0, 255),
    "AliceBlue": (240, 248, 255),
    "DarkCyan": (0, 139, 139),
    "DarkSlateGrey": (47, 79, 79),
    "GreenYellow": (173, 255, 47),
    "DarkOrchid": (153, 50, 204),
    "OliveDrab": (107, 142, 35),
    "Chartreuse": (127, 255, 0),
    "Peru": (205, 133, 63),
    "Orange": (255, 165, 0),
    "Red": (255, 0, 0),
    "Wheat": (245, 222, 179),
    "LightCyan": (224, 255, 255),
    "LightSeaGreen": (32, 178, 170),
    "BlueViolet": (138, 43, 226),
    "LightSlateGrey": (119, 136, 153),
    "Cyan": (0, 255, 255),
    "MediumPurple": (147, 112, 219),
    "MidnightBlue": (25, 25, 112),
    "FireBrick": (178, 34, 34),
    "PaleTurquoise": (175, 238, 238),
    "PaleGoldenrod": (238, 232, 170),
    "Gray": (128, 128, 128),
    "MediumSeaGreen": (60, 179, 113),
    "Moccasin": (255, 228, 181),
    "Ivory": (255, 255, 240),
    "DarkSlateBlue": (72, 61, 139),
    "Beige": (245, 245, 220),
    "Green": (0, 128, 0),
    "SlateBlue": (106, 90, 205),
    "Teal": (0, 128, 128),
    "Azure": (240, 255, 255),
    "LightSteelBlue": (176, 196, 222),
    "DimGrey": (105, 105, 105),
    "Tan": (210, 180, 140),
    "AntiqueWhite": (250, 235, 215),
    "SkyBlue": (135, 206, 235),
    "GhostWhite": (248, 248, 255),
    "MediumTurquoise": (72, 209, 204),
    "FloralWhite": (255, 250, 240),
    "LavenderBlush": (255, 240, 245),
    "SeaGreen": (46, 139, 87),
    "Lavender": (230, 230, 250),
    "BlanchedAlmond": (255, 235, 205),
    "DarkOliveGreen": (85, 107, 47),
    "DarkSeaGreen": (143, 188, 143),
    "SpringGreen": (0, 255, 127),
    "Navy": (0, 0, 128),
    "Orchid": (218, 112, 214),
    "SaddleBrown": (139, 69, 19),
    "IndianRed": (205, 92, 92),
    "Snow": (255, 250, 250),
    "SteelBlue": (70, 130, 180),
    "MediumSlateBlue": (123, 104, 238),
    "Black": (0, 0, 0),
    "LightBlue": (173, 216, 230),
    "Turquoise": (64, 224, 208),
    "MediumVioletRed": (199, 21, 133),
    "DarkViolet": (148, 0, 211),
    "DarkGray": (169, 169, 169),
    "Salmon": (250, 128, 114),
    "DarkMagenta": (139, 0, 139),
    "Tomato": (255, 99, 71),
    "WhiteSmoke": (245, 245, 245),
    "Goldenrod": (218, 165, 32),
    "MediumSpringGreen": (0, 250, 154),
    "DodgerBlue": (30, 144, 255),
    "Aqua": (0, 255, 255),
    "ForestGreen": (34, 139, 34),
    "LemonChiffon": (255, 250, 205),
    "LightSlateGray": (119, 136, 153),
    "SlateGray": (112, 128, 144),
    "LightGray": (211, 211, 211),
    "Indigo": (75, 0, 130),
    "CadetBlue": (95, 158, 160),
    "LightYellow": (255, 255, 224),
    "DarkOrange": (255, 140, 0),
    "PowderBlue": (176, 224, 230),
    "RoyalBlue": (65, 105, 225),
    "Sienna": (160, 82, 45),
    "Thistle": (216, 191, 216),
    "Lime": (0, 255, 0),
    "Seashell": (255, 245, 238),
    "DarkRed": (139, 0, 0),
    "LightSkyBlue": (135, 206, 250),
    "YellowGreen": (154, 205, 50),
    "Aquamarine": (127, 255, 212),
    "LightCoral": (240, 128, 128),
    "DarkSlateGray": (47, 79, 79),
    "Khaki": (240, 230, 140),
    "DarkGrey": (169, 169, 169),
    "BurlyWood": (222, 184, 135),
    "LightGoldenrodYellow": (250, 250, 210),
    "MediumBlue": (0, 0, 205),
    "DarkSalmon": (233, 150, 122),
    "RosyBrown": (188, 143, 143),
    "LightSalmon": (255, 160, 122),
    "PaleVioletRed": (219, 112, 147),
    "Coral": (255, 127, 80),
    "Violet": (238, 130, 238),
    "Grey": (128, 128, 128),
    "LightGreen": (144, 238, 144),
    "Linen": (250, 240, 230),
    "OrangeRed": (255, 69, 0),
    "DimGray": (105, 105, 105),
    "Maroon": (128, 0, 0),
    "LightPink": (255, 182, 193),
    "MediumAquamarine": (102, 205, 170),
    "OldLace": (253, 245, 230),
}
svgcolorslower = dict((key.lower(), value) for (key, value) in list(svgcolors.items()))


def getcolor(color):
    """Check if the given string is an RGB value, or if it is a named color.

    Parameters
    ----------
    color : str
        Color in hexadecimal format, long '#12ab9f' or short '#1af'

    Returns
    -------
    tuple
    (r, g, b, a)
        RGBA float tuple, where each value is between 0.0 and 1.0.
    """
    if color == "none":
        FreeCAD.Console.PrintMessage("Color defined as 'none', defaulting to black\n")
        return (0.0, 0.0, 0.0, 0.0)
    if color[0] == "#":
        if len(color) == 7 or len(color) == 9:  # Color string '#RRGGBB' or '#RRGGBBAA'
            r = float(int(color[1:3], 16) / 255.0)
            g = float(int(color[3:5], 16) / 255.0)
            b = float(int(color[5:7], 16) / 255.0)
            a = 1.0
            if len(color) == 9:
                a = float(int(color[7:9], 16) / 255.0)
                FreeCAD.Console.PrintMessage(f"Non standard color format #RRGGBBAA : {color}\n")
            return (r, g, b, 1 - a)
        if len(color) == 4:  # Color string '#RGB'
            # Expand the hex digits
            r = float(int(color[1], 16) * 17 / 255.0)
            g = float(int(color[2], 16) * 17 / 255.0)
            b = float(int(color[3], 16) * 17 / 255.0)
            return (r, g, b, 0.0)
    if color.lower().startswith("rgb(") or color.lower().startswith(
        "rgba("
    ):  # Color string 'rgb[a](0.12,0.23,0.3,0.0)'
        cvalues = color.lstrip("rgba(").rstrip(")").replace("%", "").split(",")
        if len(cvalues) == 3:
            a = 1.0
            if "%" in color:
                r, g, b = [int(float(cv)) / 100.0 for cv in cvalues]
            else:
                r, g, b = [int(float(cv)) / 255.0 for cv in cvalues]
        if len(cvalues) == 4:
            if "%" in color:
                r, g, b, a = [int(float(cv)) / 100.0 for cv in cvalues]
            else:
                r, g, b, a = [int(float(cv)) / 255.0 for cv in cvalues]
        return (r, g, b, 1 - a)
    # Trying named color like 'MediumAquamarine'
    v = svgcolorslower.get(color.lower())
    if v:
        r, g, b = [float(vf) / 255.0 for vf in v]
        return (r, g, b, 0.0)
    FreeCAD.Console.PrintWarning(f"Unknown color format : {color} : defaulting to black\n")
    return (0.0, 0.0, 0.0, 0.0)


def transformCopyShape(shape, m):
    """Apply transformation matrix m on given shape.

    Since OCCT 6.8.0 transformShape can be used to apply certain
    similarity transformations on shapes. This way a conversion
    to BSplines in transformGeometry can be avoided.

    @sa: Part::TopoShape::transformGeometry(), TopoShapePy::transformGeometry()
    @sa: Part::TopoShape::transformShape(), TopoShapePy::transformShape()

    Parameters
    ----------
    shape : Part::TopoShape
        A given shape
    m : Base::Matrix4D
        A transformation matrix

    Returns
    -------
    shape : Part::TopoShape
        The shape transformed by the matrix
    """
    try:
        return shape.transformShape(m, True, True)
        # Older versions of OCCT will refuse to work on
        # non-orthogonal matrices
    except Part.OCCError:
        pass
    return shape.transformGeometry(m)


def getsize(length, mode="discard", base=1):
    """Parse the length string containing number and unit.

    Parameters
    ----------
    length : str
        The length is a string, including sign, exponential notation,
        and unit: '+56215.14565E+6mm', '-23.156e-2px'.
    mode : str, optional
        One of 'discard', 'tuple', 'css90.0', 'css96.0', 'mm90.0', 'mm96.0'.
        'discard' (default), it discards the unit suffix, and extracts
            a number from the given string.
        'tuple', return number and unit as a tuple
        'css90.0', convert the unit to pixels assuming 90 dpi
        'css96.0', convert the unit to pixels assuming 96 dpi
        'mm90.0', convert the unit to millimeters assuming 90 dpi
        'mm96.0', convert the unit to millimeters assuming 96 dpi
    base : float, optional
        A base to scale the length.

    Returns
    -------
    float
        The numeric value of the length, as is, or transformed to
        millimeters or pixels.
    float, string
        A tuple with the numeric value, and the unit if `mode='tuple'`.
    """
    # Dictionaries to convert units to millimeters or pixels.
    #
    # The `em` and `ex` units are typographical units used in systems
    # like LaTeX. Here the conversion factors are arbitrarily chosen,
    # as they should depend on a specific font size used.
    #
    # The percentage factor is arbitrarily chosen, as it should depend
    # on the viewport size or for filling patterns on the bounding box.
    if mode == "mm90.0":
        tomm = {
            "": 25.4 / 90,  # default
            "px": 25.4 / 90,
            "pt": 4.0 / 3 * 25.4 / 90,
            "pc": 15 * 25.4 / 90,
            "mm": 1.0,
            "cm": 10.0,
            "in": 25.4,
            "em": 15 * 2.54 / 90,
            "ex": 10 * 2.54 / 90,
            "%": 100,
        }
    elif mode == "mm96.0":
        tomm = {
            "": 25.4 / 96,  # default
            "px": 25.4 / 96,
            "pt": 4.0 / 3 * 25.4 / 96,
            "pc": 15 * 25.4 / 96,
            "mm": 1.0,
            "cm": 10.0,
            "in": 25.4,
            "em": 15 * 2.54 / 96,
            "ex": 10 * 2.54 / 96,
            "%": 100,
        }
    elif mode == "css90.0":
        topx = {
            "": 1.0,  # default
            "px": 1.0,
            "pt": 4.0 / 3,
            "pc": 15,
            "mm": 90.0 / 25.4,
            "cm": 90.0 / 254.0,
            "in": 90,
            "em": 15,
            "ex": 10,
            "%": 100,
        }
    elif mode == "css96.0":
        topx = {
            "": 1.0,  # default
            "px": 1.0,
            "pt": 4.0 / 3,
            "pc": 15,
            "mm": 96.0 / 25.4,
            "cm": 96.0 / 254.0,
            "in": 96,
            "em": 15,
            "ex": 10,
            "%": 100,
        }

    # Extract a number from a string like '+56215.14565E+6mm'
    _num = "([-+]?[0-9]*\\.?[0-9]+([eE][-+]?[0-9]+)?)"
    _unit = "(px|pt|pc|mm|cm|in|em|ex|%)?"
    _full_num = _num + _unit
    number, exponent, unit = re.findall(_full_num, length)[0]
    if mode == "discard":
        return float(number)
    elif mode == "tuple":
        return float(number), unit
    elif mode == "isabsolute":
        return unit in ("mm", "cm", "in", "px", "pt")
    elif mode == "mm96.0" or mode == "mm90.0":
        return float(number) * tomm[unit]
    elif mode == "css96.0" or mode == "css90.0":
        if unit != "%":
            return float(number) * topx[unit]
        else:
            return float(number) * base


def getrgb(color):
    """Return an RGB hexadecimal string '#00aaff' from a FreeCAD color.

    Parameters
    ----------
    color : Base::Color::Color
        FreeCAD color.

    Returns
    -------
    str
        The hexadecimal string representation of the color '#00aaff'.
    """
    r = str(hex(int(color[0] * 255)))[2:].zfill(2)
    g = str(hex(int(color[1] * 255)))[2:].zfill(2)
    b = str(hex(int(color[2] * 255)))[2:].zfill(2)
    return "#" + r + g + b


class svgHandler(xml.sax.ContentHandler):
    """Parse SVG files and create FreeCAD objects."""

    def __init__(self):
        super().__init__()
        """Retrieve Draft parameters and initialize."""
        self.style = params.get_param("svgstyle")
        self.disableUnitScaling = params.get_param("svgDisableUnitScaling")
        self.make_cuts = params.get_param("svgMakeCuts")
        self.add_wire_for_invalid_face = params.get_param("svgAddWireForInvalidFace")
        self.count = 0
        self.transform = None
        self.grouptransform = []
        self.groupstyles = []
        self.lastdim = None
        self.viewbox = None
        self.svgdpi = 1.0

        global Part

        if gui and draftui:
            r = float(draftui.color.red() / 255.0)
            g = float(draftui.color.green() / 255.0)
            b = float(draftui.color.blue() / 255.0)
            rf = float(draftui.facecolor.red() / 255.0)
            gf = float(draftui.facecolor.green() / 255.0)
            bf = float(draftui.facecolor.blue() / 255.0)
            self.width_default = float(draftui.linewidth)
        else:
            self.width_default = float(params.get_param_view("DefaultShapeLineWidth"))
            r, g, b, _ = utils.get_rgba_tuple(params.get_param_view("DefaultShapeLineColor"))
            rf, gf, bf, _ = utils.get_rgba_tuple(params.get_param_view("DefaultShapeColor"))
        self.fill_default = (rf, gf, bf, 0.0)
        self.color_default = (r, g, b, 0.0)

    def format(self, obj):
        """Apply styles to the object if the graphical interface is up."""
        if FreeCAD.GuiUp:
            v = obj.ViewObject
            if self.color:
                v.LineColor = self.color
            if self.width:
                v.LineWidth = self.width
            if self.fill:
                v.ShapeColor = self.fill

    def __addFaceToDoc(self, named_face):
        """Create a named document object from a name/face tuple

        Parameters
        ----------
        named_face : name : str, face : Part.Face
                     The Face/Wire to add, and its name
        """
        name, face = named_face
        if not face:
            return

        face = self.applyTrans(face)
        obj = self.doc.addObject("Part::Feature", name)
        obj.Shape = face
        self.format(obj)

    def startElement(self, name, attrs):
        """Re-organize data into a nice clean dictionary.

        Parameters
        ----------
        name : str
            Name of the element: 'path', 'rect', 'line', 'polyline',
            'polygon', 'ellipse', 'circle', 'text', 'tspan', 'symbol'
        attrs : iterable
            Dictionary of content of the elements
        """
        self.count += 1
        precision = svg_precision()

        _msg("processing element {0}: {1}".format(self.count, name))
        _msg("existing group transform: {}".format(self.grouptransform))
        _msg("existing group style: {}".format(self.groupstyles))

        data = {}
        for keyword, content in list(attrs.items()):
            # print(keyword, content)
            if keyword != "style":
                content = content.replace(",", " ")
                content = content.split()
            # print(keyword, content)
            data[keyword] = content

        # If it's the first element, which is <svg>,
        # check if the file is created by Inkscape, and its version,
        # in order to consider some attributes of the SVG file.
        if self.count == 1 and name == "svg":
            if "inkscape:version" in data:
                inks_doc_name = attrs.getValue("sodipodi:docname")
                inks_full_ver = attrs.getValue("inkscape:version")
                inks_ver_pars = re.search("\\d+\\.\\d+", inks_full_ver)
                if inks_ver_pars is not None:
                    inks_ver_f = float(inks_ver_pars.group(0))
                else:
                    inks_ver_f = 99.99
                # Inkscape before 0.92 used 90 dpi as resolution
                # Newer versions use 96 dpi
                if inks_ver_f < 0.92:
                    self.svgdpi = 90.0
                else:
                    self.svgdpi = 96.0
            if "inkscape:version" not in data:
                # exact scaling is calculated later below. Here we just want
                # to skip the DPI dialog if a unit is specified in the viewbox
                if "width" in data and "mm" in attrs.getValue("width"):
                    self.svgdpi = 96.0
                elif "width" in data and "in" in attrs.getValue("width"):
                    self.svgdpi = 96.0
                elif "width" in data and "cm" in attrs.getValue("width"):
                    self.svgdpi = 96.0
                else:
                    _inf = (
                        "This SVG file does not appear to have been produced "
                        "by Inkscape. If it does not contain absolute units "
                        "then a DPI setting will be used."
                    )
                    _qst = (
                        "Do you wish to use 96 dpi? Choosing 'No' "
                        "will use the older standard 90 dpi."
                    )
                    if FreeCAD.GuiUp:
                        msgBox = QtWidgets.QMessageBox()
                        msgBox.setText(translate("ImportSVG", _inf))
                        msgBox.setInformativeText(translate("ImportSVG", _qst))
                        msgBox.setStandardButtons(
                            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No
                        )
                        msgBox.setDefaultButton(QtWidgets.QMessageBox.No)
                        ret = msgBox.exec_()
                        if ret == QtWidgets.QMessageBox.Yes:
                            self.svgdpi = 96.0
                        else:
                            self.svgdpi = 90.0
                        if ret:
                            _msg(translate("ImportSVG", _inf))
                            _msg(translate("ImportSVG", _qst))
                            _msg("*** User specified {} " "dpi ***".format(self.svgdpi))
                    else:
                        self.svgdpi = 96.0
                        _msg(_inf)
                        _msg("*** Assuming {} dpi ***".format(self.svgdpi))
            if self.svgdpi == 1.0:
                _wrn(
                    "This SVG file (" + inks_doc_name + ") "
                    "has an unrecognised format which means "
                    "the dpi could not be determined; "
                    "assuming 96 dpi"
                )
                self.svgdpi = 96.0

        if "style" in data:
            if not data["style"]:
                # Empty style attribute stops inheriting from parent
                pass
            else:
                content = data["style"].replace(" ", "")
                content = content.split(";")
                for i in content:
                    pair = i.split(":")
                    if len(pair) > 1:
                        data[pair[0]] = pair[1]

        for k in ["x", "y", "x1", "y1", "x2", "y2", "r", "rx", "ry", "cx", "cy", "width", "height"]:
            if k in data:
                data[k] = getsize(data[k][0], "css" + str(self.svgdpi))

        for k in ["fill", "stroke", "stroke-width", "font-size"]:
            if k in data:
                if isinstance(data[k], list):
                    if data[k][0].lower().startswith("rgb("):
                        data[k] = ",".join(data[k])
                    else:
                        data[k] = data[k][0]

        # Extract style info
        self.fill = None
        self.color = None
        self.width = None
        self.text = None

        if name == "svg":
            m = FreeCAD.Matrix()
            if not self.disableUnitScaling:
                if "width" in data and "height" in data and "viewBox" in data:
                    if len(self.grouptransform) == 0:
                        unitmode = "mm" + str(self.svgdpi)
                    else:
                        # nested svg element
                        unitmode = "css" + str(self.svgdpi)
                    vbw = round(getsize(data["viewBox"][2], "discard"), precision)
                    vbh = round(getsize(data["viewBox"][3], "discard"), precision)
                    abw = round(getsize(attrs.getValue("width"), unitmode), precision)
                    abh = round(getsize(attrs.getValue("height"), unitmode), precision)
                    self.viewbox = (vbw, vbh)
                    sx = abw / vbw
                    sy = abh / vbh
                    preserve_ar = " ".join(data.get("preserveAspectRatio", [])).lower()
                    if preserve_ar.startswith("none"):
                        m.scale(Vector(sx, sy, 1))
                        if sx != sy:
                            _wrn(
                                "Non-uniform scaling with probably degenerating "
                                + "effects on Edges. ({} vs. {}).".format(sx, sy)
                            )

                    else:
                        # preserve aspect ratio - svg default is 'x/y-mid meet'
                        if preserve_ar.endswith("slice"):
                            sxy = max(sx, sy)
                        else:
                            sxy = min(sx, sy)
                        m.scale(Vector(sxy, sxy, 1))
                elif len(self.grouptransform) == 0:
                    # fallback to current dpi
                    m.scale(Vector(25.4 / self.svgdpi, 25.4 / self.svgdpi, 1))
            self.grouptransform.append(m)
        if "fill" in data:
            if data["fill"] != "none":
                self.fill = getcolor(data["fill"])
        if "stroke" in data:
            if data["stroke"] != "none":
                self.color = getcolor(data["stroke"])
        if "stroke-width" in data:
            if data["stroke-width"] != "none":
                self.width = getsize(data["stroke-width"], "css" + str(self.svgdpi))
        if "transform" in data:
            m = self.getMatrix(attrs.getValue("transform"))
        else:
            m = FreeCAD.Matrix()
        if name == "g" or name == "a":
            self.grouptransform.append(m)
        elif name == "freecad:used":
            # use tag acts as g tag but has x,y attribute
            x = data.get("x", 0)
            y = data.get("y", 0)
            if x != 0 or y != 0:
                xy = FreeCAD.Matrix()
                xy.move(Vector(x, -y, 0))
                m = m.multiply(xy)
            self.grouptransform.append(m)
        elif "transform" in data:
            self.transform = m

        if self.style == 0:
            if self.fill is not None:
                self.fill = self.fill_default
            self.color = self.color_default
            self.width = self.width_default

        # apply group styles
        if name == "g" or name == "a" or name == "freecad:used":
            self.groupstyles.append([self.fill, self.color, self.width])
        if self.fill is None:
            if "fill" not in data:
                # do not override fill if this item has specifically set a none fill
                for groupstyle in reversed(self.groupstyles):
                    if groupstyle[0] is not None:
                        self.fill = groupstyle[0]
                        break
                if self.fill is None:
                    # svg fill default is Black
                    self.fill = getcolor("Black")
        if self.color is None:
            for groupstyle in reversed(self.groupstyles):
                if groupstyle[1] is not None:
                    self.color = groupstyle[1]
                    break
        if self.width is None:
            for groupstyle in reversed(self.groupstyles):
                if groupstyle[2] is not None:
                    self.width = groupstyle[2]
                    break

        pathname = None
        if "id" in data:
            pathname = data["id"][0]
            _msg("name: {}".format(pathname))

        # Process paths
        if name == "path":
            if not pathname:
                pathname = "Path"
            _msg("data: {}".format(data))

            if "freecad:basepoint1" in data:
                p1 = data["freecad:basepoint1"]
                p1 = Vector(float(p1[0]), -float(p1[1]), 0)
                p2 = data["freecad:basepoint2"]
                p2 = Vector(float(p2[0]), -float(p2[1]), 0)
                p3 = data["freecad:dimpoint"]
                p3 = Vector(float(p3[0]), -float(p3[1]), 0)
                obj = Draft.make_dimension(p1, p2, p3)
                self.applyTrans(obj)
                self.format(obj)
                self.lastdim = obj
                data["d"] = []

            if "d" in data:
                svgPath = SvgPathParser(data, pathname)
                svgPath.parse()
                svgPath.create_faces(self.fill, self.add_wire_for_invalid_face)
                if self.make_cuts:
                    svgPath.doCuts()
                shapes = svgPath.getShapeList()
                for named_shape in shapes:
                    self.__addFaceToDoc(named_shape)

        # Process rects
        if name == "rect":
            if not pathname:
                pathname = "Rectangle"
            edges = []
            if "x" not in data:
                data["x"] = 0
            if "y" not in data:
                data["y"] = 0
            # Negative values are invalid
            _precision = 10 ** (-precision)
            if ("rx" not in data or data["rx"] < _precision) and (
                "ry" not in data or data["ry"] < _precision
            ):
                # if True:
                p1 = Vector(data["x"], -data["y"], 0)
                p2 = Vector(data["x"] + data["width"], -data["y"], 0)
                p3 = Vector(data["x"] + data["width"], -data["y"] - data["height"], 0)
                p4 = Vector(data["x"], -data["y"] - data["height"], 0)
                edges.append(Part.LineSegment(p1, p2).toShape())
                edges.append(Part.LineSegment(p2, p3).toShape())
                edges.append(Part.LineSegment(p3, p4).toShape())
                edges.append(Part.LineSegment(p4, p1).toShape())
            else:
                # rounded edges
                rx = data.get("rx")
                ry = data.get("ry") or rx
                rx = rx or ry
                if rx > 2 * data["width"]:
                    rx = data["width"] / 2.0
                if ry > 2 * data["height"]:
                    ry = data["height"] / 2.0

                p1 = Vector(data["x"] + rx, -data["y"] - data["height"] + ry, 0)
                p2 = Vector(data["x"] + data["width"] - rx, -data["y"] - data["height"] + ry, 0)
                p3 = Vector(data["x"] + data["width"] - rx, -data["y"] - ry, 0)
                p4 = Vector(data["x"] + rx, -data["y"] - ry, 0)

                if rx < 0 or ry < 0:
                    _wrn("Warning: 'rx' or 'ry' is negative, " "check the SVG file")

                if rx >= ry:
                    e = Part.Ellipse(Vector(), rx, ry)
                    e1a = Part.Arc(e, math.radians(180), math.radians(270))
                    e2a = Part.Arc(e, math.radians(270), math.radians(360))
                    e3a = Part.Arc(e, math.radians(0), math.radians(90))
                    e4a = Part.Arc(e, math.radians(90), math.radians(180))
                    m = FreeCAD.Matrix()
                else:
                    e = Part.Ellipse(Vector(), ry, rx)
                    e1a = Part.Arc(e, math.radians(90), math.radians(180))
                    e2a = Part.Arc(e, math.radians(180), math.radians(270))
                    e3a = Part.Arc(e, math.radians(270), math.radians(360))
                    e4a = Part.Arc(e, math.radians(0), math.radians(90))
                    # rotate +90 degrees
                    m = FreeCAD.Matrix(0, -1, 0, 0, 1, 0)
                esh = []
                for arc, point in ((e1a, p1), (e2a, p2), (e3a, p3), (e4a, p4)):
                    m1 = FreeCAD.Matrix(m)
                    m1.move(point)
                    arc.transform(m1)
                    esh.append(arc.toShape())
                for esh1, esh2 in zip(esh[-1:] + esh[:-1], esh):
                    p1 = esh1.Vertexes[-1].Point
                    p2 = esh2.Vertexes[0].Point
                    if not equals(p1, p2, precision):
                        # straight segments
                        _sh = Part.LineSegment(p1, p2).toShape()
                        edges.append(_sh)
                    # elliptical segments
                    edges.append(esh2)
            sh = Part.Wire(edges)
            if self.fill:
                sh = Part.Face(sh)
            sh = self.applyTrans(sh)
            obj = self.doc.addObject("Part::Feature", pathname)
            obj.Shape = sh
            self.format(obj)

        # Process lines
        if name == "line":
            if not pathname:
                pathname = "Line"
            p1 = Vector(data["x1"], -data["y1"], 0)
            p2 = Vector(data["x2"], -data["y2"], 0)
            sh = Part.LineSegment(p1, p2).toShape()
            sh = self.applyTrans(sh)
            obj = self.doc.addObject("Part::Feature", pathname)
            obj.Shape = sh
            self.format(obj)

        # Process polylines and polygons
        if name == "polyline" or name == "polygon":
            # A simpler implementation would be
            # _p = zip(points[0::2], points[1::2])
            # sh = Part.makePolygon([Vector(svgx,
            #                               -svgy,
            #                               0) for svgx, svgy in _p])
            #
            # but it would be more difficult to search for duplicate
            # points beforehand.
            if not pathname:
                pathname = "Polyline"
            points = [float(d) for d in data["points"]]
            lenpoints = len(points)
            if lenpoints >= 4 and lenpoints % 2 == 0:
                lastvec = Vector(points[0], -points[1], 0)
                path = []
                if name == "polygon":
                    points = points + points[:2]  # emulate closepath
                for svgx, svgy in zip(points[2::2], points[3::2]):
                    currentvec = Vector(svgx, -svgy, 0)
                    if not equals(lastvec, currentvec, precision):
                        seg = Part.LineSegment(lastvec, currentvec).toShape()
                        # print("polyline seg ", lastvec, currentvec)
                        lastvec = currentvec
                        path.append(seg)
                if path:
                    sh = Part.Wire(path)
                    if self.fill and sh.isClosed():
                        sh = Part.Face(sh)
                    sh = self.applyTrans(sh)
                    obj = self.doc.addObject("Part::Feature", pathname)
                    obj.Shape = sh
                    self.format(obj)

        # Process ellipses
        if name == "ellipse":
            if not pathname:
                pathname = "Ellipse"
            c = Vector(data.get("cx", 0), -data.get("cy", 0), 0)
            rx = data["rx"]
            ry = data["ry"]

            if rx < 0 or ry < 0:
                _wrn("Warning: 'rx' or 'ry' is negative, check the SVG file")

            if rx > ry:
                sh = Part.Ellipse(c, rx, ry).toShape()
            else:
                sh = Part.Ellipse(c, ry, rx).toShape()
                sh.rotate(c, Vector(0, 0, 1), 90)
            if self.fill:
                sh = Part.Wire([sh])
                sh = Part.Face(sh)
            sh = self.applyTrans(sh)
            obj = self.doc.addObject("Part::Feature", pathname)
            obj.Shape = sh
            self.format(obj)

        # Process circles
        if name == "circle" and "freecad:skip" not in data:
            if not pathname:
                pathname = "Circle"
            c = Vector(data.get("cx", 0), -data.get("cy", 0), 0)
            r = data["r"]
            sh = Part.makeCircle(r)
            if self.fill:
                sh = Part.Wire([sh])
                sh = Part.Face(sh)
            sh.translate(c)
            sh = self.applyTrans(sh)
            obj = self.doc.addObject("Part::Feature", pathname)
            obj.Shape = sh
            self.format(obj)

        # Process texts
        if name in ["text", "tspan"]:
            if "freecad:skip" not in data:
                _msg("processing a text")
                if "x" in data:
                    self.x = data["x"]
                else:
                    self.x = 0
                if "y" in data:
                    self.y = data["y"]
                else:
                    self.y = 0
                if "font-size" in data:
                    if data["font-size"] != "none":
                        self.text = getsize(data["font-size"], "css" + str(self.svgdpi))
                else:
                    self.text = 1
            else:
                if self.lastdim:
                    _font_size = int(getsize(data["font-size"]))
                    self.lastdim.ViewObject.FontSize = _font_size

        _msg("done processing element {}".format(self.count))

    # startElement()

    def characters(self, content):
        """Read characters from the given string."""
        if self.text:
            _msg("reading characters {}".format(content))
            obj = self.doc.addObject("App::Annotation", "Text")
            # use ignore to not break import if char is not found in latin1
            obj.LabelText = content.encode("latin1", "ignore")
            vec = Vector(self.x, -self.y, 0)
            if self.transform:
                vec = self.translateVec(vec, self.transform)
                # print("own transform: ", self.transform, vec)
            for transform in self.grouptransform[::-1]:
                # vec = self.translateVec(vec, transform)
                vec = transform.multiply(vec)
            # print("applying vector: ", vec)
            obj.Position = vec
            if FreeCAD.GuiUp:
                obj.ViewObject.FontSize = int(self.text)
                if self.fill:
                    obj.ViewObject.TextColor = self.fill
                else:
                    obj.ViewObject.TextColor = (0.0, 0.0, 0.0, 0.0)

    def endElement(self, name):
        """Finish processing the element indicated by the name.

        Parameters
        ----------
        name : str
            The name of the element
        """
        if name not in ["tspan"]:
            self.transform = None
            self.text = None
        if name == "g" or name == "a" or name == "svg" or name == "freecad:used":
            _msg("closing group")
            self.grouptransform.pop()
            if self.groupstyles:
                self.groupstyles.pop()

    def applyTrans(self, sh):
        """Apply transformation to the shape and return the new shape.

        Parameters
        ----------
        sh : Part.Shape or Draft.Dimension
            Object to be transformed
        """
        if isinstance(sh, Part.Shape) or isinstance(sh, Part.Wire):
            if self.transform:
                sh = transformCopyShape(sh, self.transform)
            for transform in self.grouptransform[::-1]:
                sh = transformCopyShape(sh, transform)
            return sh
        elif Draft.getType(sh) in ["Dimension", "LinearDimension"]:
            pts = []
            for p in [sh.Start, sh.End, sh.Dimline]:
                cp = Vector(p)
                if self.transform:
                    cp = self.transform.multiply(cp)
                for transform in self.grouptransform[::-1]:
                    cp = transform.multiply(cp)
                pts.append(cp)
            sh.Start = pts[0]
            sh.End = pts[1]
            sh.Dimline = pts[2]

    def translateVec(self, vec, mat):
        """Translate (move) a point or vector by a matrix.

        Parameters
        ----------
        vec : Base::Vector3
            The original vector
        mat : Base::Matrix4D
            The translation matrix, from which only the elements 14, 24, 34
            are used.
        """
        v = Vector(mat.A14, mat.A24, mat.A34)
        return vec.add(v)

    def getMatrix(self, tr):
        """Return a FreeCAD matrix from an SVG transform attribute.

        Parameters
        ----------
        tr : str
            The type of transform: 'matrix', 'translate', 'scale',
            'rotate', 'skewX', 'skewY' and its value

        Returns
        -------
        Base::Matrix4D
            The translated matrix.
        """
        _op = "(matrix|translate|scale|rotate|skewX|skewY)"
        _val = "\\((.*?)\\)"
        _transf = _op + "\\s*?" + _val
        transformre = re.compile(_transf, re.DOTALL)
        m = FreeCAD.Matrix()
        for transformation, arguments in reversed(transformre.findall(tr)):
            _args_rep = arguments.replace(",", " ").split()
            argsplit = [float(arg) for arg in _args_rep]
            # m.multiply(FreeCAD.Matrix(1, 0, 0, 0, 0, -1))
            # print('%s:%s %s %d' % (transformation, arguments,
            #                        argsplit, len(argsplit)))
            if transformation == "translate":
                tx = argsplit[0]
                ty = argsplit[1] if len(argsplit) > 1 else 0.0
                m.move(Vector(tx, -ty, 0))
            elif transformation == "scale":
                sx = argsplit[0]
                sy = argsplit[1] if len(argsplit) > 1 else sx
                m.scale(Vector(sx, sy, 1))
            elif transformation == "rotate":
                cx = 0
                cy = 0
                angle = argsplit[0]
                if len(argsplit) >= 3:
                    # Rotate around a non-origin centerpoint (note: SVG y axis is opposite FreeCAD y axis)
                    cx = argsplit[1]
                    cy = argsplit[2]
                    m.move(Vector(-cx, cy, 0))  # Reposition for rotation
                # Mirroring one axis is equal to changing the direction
                # of rotation
                m.rotateZ(math.radians(-angle))
                if len(argsplit) >= 3:
                    m.move(Vector(cx, -cy, 0))  # Reverse repositioning
            elif transformation == "skewX":
                _m = FreeCAD.Matrix(1, -math.tan(math.radians(argsplit[0])))
                m = m.multiply(_m)
            elif transformation == "skewY":
                _m = FreeCAD.Matrix(1, 0, 0, 0, -math.tan(math.radians(argsplit[0])))
                m = m.multiply(_m)
            elif transformation == "matrix":
                # transformation matrix:
                #    FreeCAD                 SVG
                # (+A -C +0 +E)           (A C 0 E)
                # (-B +D -0 -F)  = (-Y) * (B D 0 F) * (-Y)
                # (+0 -0 +1 +0)           (0 0 1 0)
                # (+0 -0 +0 +1)           (0 0 0 1)
                #
                # Put the first two rows of the matrix
                # fmt: off
                _m = FreeCAD.Matrix(argsplit[0], -argsplit[2],
                                    0, argsplit[4],
                                    -argsplit[1], argsplit[3],
                                    0, -argsplit[5])
                # fmt: on
                m = m.multiply(_m)
            # else:
            #    print('SKIPPED %s' % transformation)
            # print("m = ", m)
        # print("generating transformation: ", m)
        return m

    # getMatrix


# class svgHandler


def getContents(filename, tag, stringmode=False):
    """Get the contents of all occurrences of the given tag in the file.

    Parameters
    ----------
    filename : str
        A filename to scan for tags.
    tag : str
        An SVG tag to find inside a file, for example, `some`
        in <some id="12">information</some>
    stringmode : bool, optional
        The default is False.
        If False, `filename` is a path to a file.
        If True, `filename` is already a pointer to an open file.

    Returns
    -------
    dict
        A dictionary with tagids and the information associated with that id
        results[tagid] = information
    """
    result = {}
    if stringmode:
        contents = filename
    else:
        # Use the native Python open which was saved as `pyopen`
        f = pyopen(filename)
        contents = f.read()
        f.close()

    # Replace the newline character with a string
    # so that it's easiert to parse; later on the newline character
    # will be restored
    contents = contents.replace("\n", "_linebreak")
    searchpat = "<" + tag + ".*?</" + tag + ">"
    tags = re.findall(searchpat, contents)
    for t in tags:
        tagid = re.findall(r'id="(.*?)"', t)
        if tagid:
            tagid = tagid[0]
        else:
            tagid = "none"
        res = t.replace("_linebreak", "\n")
        result[tagid] = res
    return result


def open(filename):
    """Open filename and parse using the svgHandler().

    Parameters
    ----------
    filename : str
        The path to the filename to be opened.

    Returns
    -------
    App::Document
        The new FreeCAD document object created, with the parsed information.
    """
    docname = os.path.split(filename)[1]
    doc = FreeCAD.newDocument(docname)
    doc.Label = docname[:-4]

    # Set up the parser
    parser = xml.sax.make_parser()
    parser.setFeature(xml.sax.handler.feature_external_ges, False)
    handler = svgHandler()
    parser.setContentHandler(handler)
    parser._cont_handler.doc = doc

    # preprocess file to replace use tag to it's referenced object
    new_svg_content = replace_use_with_reference(filename)
    xml.sax.parseString(new_svg_content, handler)
    doc.recompute()
    return doc


def insert(filename, docname):
    """Get an active document and parse using the svgHandler().

    If no document exist, it is created.

    Parameters
    ----------
    filename : str
        The path to the filename to be opened.
    docname : str
        The name of the active App::Document if one exists, or
        of the new one created.

    Returns
    -------
    App::Document
        The active FreeCAD document, or the document created if none exists,
        with the parsed information.
    """
    try:
        doc = FreeCAD.getDocument(docname)
    except NameError:
        doc = FreeCAD.newDocument(docname)
    FreeCAD.ActiveDocument = doc

    # Set up the parser
    parser = xml.sax.make_parser()
    parser.setFeature(xml.sax.handler.feature_external_ges, False)
    handler = svgHandler()
    parser.setContentHandler(handler)
    parser._cont_handler.doc = doc

    # preprocess file to replace use tag to it's referenced object
    new_svg_content = replace_use_with_reference(filename)
    xml.sax.parseString(new_svg_content, handler)
    doc.recompute()


def export(exportList, filename):
    """Export the SVG file with a given list of objects.

    The objects must be derived from Part::Feature, in order to be processed
    and exported.

    Parameters
    ----------
    exportList : list
        List of document objects to export.
    filename : str
        Path to the new file.

    Returns
    -------
    None
        If `exportList` doesn't have shapes to export.
    """
    svg_export_style = params.get_param("svg_export_style")
    if svg_export_style != 0 and svg_export_style != 1:
        _msg(translate("ImportSVG", "Unknown SVG export style, switching to Translated"))
        svg_export_style = 0

    tmp = []
    hidden_doc = None
    base_sketch_pla = None  # Placement of the 1st sketch.
    for obj in exportList:
        if obj.isDerivedFrom("Sketcher::SketchObject"):
            if hidden_doc is None:
                hidden_doc = FreeCAD.newDocument(name="hidden", hidden=True, temp=True)
                base_sketch_pla = obj.Placement
            sh = Part.Compound()
            sh.Placement = base_sketch_pla
            sh.add(obj.Shape.copy())
            sh.transformShape(base_sketch_pla.inverse().Matrix)
            new = hidden_doc.addObject("Part::Part2DObjectPython")
            new.Shape = sh
            if FreeCAD.GuiUp:
                for attr in ("DrawStyle", "LineColor", "LineWidth"):
                    setattr(new.ViewObject, attr, getattr(obj.ViewObject, attr))
            tmp.append(new)
        else:
            tmp.append(obj)
    exportList = tmp

    # Determine the size of the page by adding the bounding boxes
    # of all shapes
    bb = FreeCAD.BoundBox()
    for obj in exportList:
        if hasattr(obj, "Shape") and obj.Shape and obj.Shape.BoundBox.isValid():
            bb.add(obj.Shape.BoundBox)
        else:
            # if Draft.get_type(obj) in ("Text", "LinearDimension", ...)
            _wrn("'{}': no Shape, " "calculate manual bounding box".format(obj.Label))
            bb.add(Draft.get_bbox(obj))

    if not bb.isValid():
        _err(
            translate(
                "ImportSVG", "The export list contains no object " "with a valid bounding box"
            )
        )
        return

    minx = bb.XMin
    maxx = bb.XMax
    miny = bb.YMin
    maxy = bb.YMax

    if svg_export_style == 0:
        # translated-style exports get a bit of a margin
        margin = (maxx - minx) * 0.01
    else:
        # raw-style exports get no margin
        margin = 0

    minx -= margin
    maxx += margin
    miny -= margin
    maxy += margin
    sizex = maxx - minx
    sizey = maxy - miny
    miny += margin

    # Use the native Python open which was saved as `pyopen`
    svg = pyopen(filename, "w")

    # Write header.
    # We specify the SVG width and height in FreeCAD's physical units (mm),
    # and specify the viewBox so that user units maps one-to-one to mm.
    svg.write('<?xml version="1.0"?>\n')
    svg.write('<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"')
    svg.write(' "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">\n')
    svg.write("<svg")
    svg.write(' width="' + str(sizex) + 'mm" height="' + str(sizey) + 'mm"')
    if svg_export_style == 0:
        # translated-style exports have the viewbox starting at X=0, Y=0
        svg.write(' viewBox="0 0 ' + str(sizex) + " " + str(sizey) + '"')
    else:
        # Raw-style exports have the viewbox starting at X=xmin, Y=-ymax
        # We need the negative Y here because SVG is upside down, and we
        # flip the sketch right-way up with a scale later
        svg.write(' viewBox="%f %f %f %f"' % (minx, -maxy, sizex, sizey))

    svg.write(' xmlns="http://www.w3.org/2000/svg" version="1.1"')
    svg.write(">\n")

    # Write paths
    for ob in exportList:
        if svg_export_style == 0:
            # translated-style exports have the entire sketch translated
            # to fit in the X>0, Y>0 quadrant
            # svg.write('<g transform="translate('
            #           + str(-minx) + ',' + str(-miny + 2*margin)
            #           + ') scale(1,-1)">\n')
            svg.write(
                '<g id="%s" transform="translate(%f,%f) ' 'scale(1,-1)">\n' % (ob.Name, -minx, maxy)
            )
        else:
            # raw-style exports do not translate the sketch
            svg.write('<g id="%s" transform="scale(1,-1)">\n' % ob.Name)

        svg.write(Draft.get_svg(ob, override=False))
        _label_enc = str(ob.Label.encode("utf8"))
        _label = _label_enc.replace("<", "&lt;").replace(">", "&gt;")
        # replace('"', "&quot;")
        svg.write("<title>%s</title>\n" % _label)
        svg.write("</g>\n")

    # Close the file
    svg.write("</svg>")
    svg.close()
    if hidden_doc is not None:
        try:
            App.closeDocument(hidden_doc.Name)
        except:
            pass


# function to replace use tag to it's referenced object
def replace_use_with_reference(file_path):
    # function that replace use tag to freecad:used
    def register_svg_namespaces(svg_content):
        # register namespaces
        xmlns_attrs = re.findall(r'\s+xmlns(?::([a-zA-Z0-9_]+))?="([^"]+)"', svg_content)
        for prefix, uri in xmlns_attrs:
            ns_prefix = "" if prefix is None or prefix == "svg" else prefix
            ET.register_namespace(ns_prefix, uri)

    def replace_use(element, tree):
        while True:
            uses = element.findall(".//{http://www.w3.org/2000/svg}use")
            if uses == []:
                break
            # create parent map
            parent_map = {child: parent for parent in tree.iter() for child in parent}
            for use in uses:
                parent = parent_map[use]
                href = use.attrib.get("href", "")
                # if href is empty, try to get xlink:href.
                if not href:
                    href = use.attrib.get("{http://www.w3.org/1999/xlink}href", "")
                if href.startswith("#"):
                    ref_id = href[1:]
                    ref_element = id_map.get(ref_id)
                    if ref_element is not None:
                        # defs tag could not be referenced by use tag.
                        if ref_element.tag.endswith("defs"):
                            continue
                        # make new element named freecad:used because use tag may have it own transform.
                        new_element = ET.Element("freecad:used")
                        for attr in use.attrib:
                            # copy attribute to new one except href attribute
                            if (
                                attr not in {"href", "{http://www.w3.org/1999/xlink}href"}
                                and attr not in new_element.attrib
                            ):
                                new_element.set(attr, use.attrib[attr])
                        ref_element = deepcopy(ref_element)
                        # change referenced symbol tag to g tag, because symbol tag will be ignored when importing.
                        if ref_element.tag.endswith("symbol"):
                            ref_element.tag = "g"
                        # remove id from referenced element.(without this multiple same id problem)
                        if "id" in ref_element.attrib:
                            del ref_element.attrib["id"]
                        for child in list(ref_element):
                            # remove id from child of referenced element.(without this multiple same id problem)
                            if "id" in child.attrib:
                                del child.attrib["id"]
                        new_element.append(ref_element)
                        # replace use tag by freecad:used tag.
                        parent.append(new_element)
                # remove use when referenced element is not found.
                parent.remove(use)
        # now all use tag processd
        # remove symbol and defs tag from tree.
        parent_map = {child: parent for parent in tree.iter() for child in parent}
        symbols = element.findall(".//{http://www.w3.org/2000/svg}symbol")
        for symbol in symbols:
            parent = parent_map[symbol]
            parent.remove(symbol)
        deftags = element.findall(".//{http://www.w3.org/2000/svg}defs")
        for deftag in deftags:
            parent = parent_map[deftag]
            parent.remove(deftag)

    # open file and read
    svg_content = pyopen(file_path).read()
    # register namespace before parsing
    register_svg_namespaces(svg_content)
    # parse as xml.
    tree = ET.ElementTree(ET.fromstring(svg_content))
    root = tree.getroot()

    # create id dictionary.
    id_map = {}
    for elem in root.findall(".//*[@id]"):
        id_map[elem.attrib["id"]] = elem

    replace_use(root, tree)

    # return tree as xml string with namespace declaration.
    return ET.tostring(root, encoding="unicode", xml_declaration=True)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2019 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

## \defgroup OFFLINERENDERINGUTILS OfflineRenderingUtils
#  \ingroup UTILITIES
#  \brief Utility functions to work with FreeCAD files in console mode
#
# Offline rendering utilities

## \addtogroup OFFLINERENDERINGUTILS
#  @{

"""@package docstring
OfflineRenderingUtils - Utilities to help producing files with colors from FreeCAD in non-GUI mode


Example usage:

The examples below extract the colors from an existing file, but you can also create your own
while working, as a simple dictionary of "objectName":(r,g,b) pairs (r, g, b as floats)

import os
import sys
import FreeCAD
import OfflineRenderingUtils

# build full filepaths

freecadFile = os.path.join(os.getcwd(),"testfile.FCStd")
baseFileName = os.path.splitext(freecadFile)[0]

# open FreeCAD file

doc = FreeCAD.open(freecadFile)

# build color dict

# setting nodiffuse=True (optional) discards per-face colors, and only sets one color per object
# only the STEP exporter accepts per-face colors. The others would consider the first color in the per-face colors list as
# the object color, which might be not what we want, so it's best to turn it off here.

colors = OfflineRenderingUtils.getColors(freecadFile,nodiffuse=True)

# get the camera data from the file (used in some functions below)

camera = OfflineRenderingUtils.getCamera(freecadFile)

# export to OBJ

from importers import importOBJ
importOBJ.export(doc.Objects,baseFileName+".obj",colors=colors)

# export to DAE

from importers import importDAE
importDAE.export(doc.Objects,baseFileName+".dae",colors=colors)

# export to IFC

from importers import importIFC
importIFC.export(doc.Objects,baseFileName+".ifc",colors=colors)

# export to STEP

# The STEP exporter accepts different kind of data than the above
# exporters. Use the function below to reformat it the proper way

stepdata = OfflineRenderingUtils.getStepData(doc.Objects,colors)
import Import
Import.export(stepdata,baseFileName+".stp")

# export to PNG

scene = OfflineRenderingUtils.buildScene(doc.Objects,colors)
OfflineRenderingUtils.render(baseFileName+".png",scene,camera,width=800,height=600)

# view the scene in a standalone coin viewer

OfflineRenderingUtils.viewer(scene)

# file saving

OfflineRenderingUtils.save(doc,filename=baseFileName+"_exported.FCStd",colors=colors,camera=camera)
"""

import binascii
import inspect
import os
import sys
import tempfile
import xml.sax
import zipfile


class FreeCADGuiHandler(xml.sax.ContentHandler):
    "A XML handler to process the FreeCAD GUI xml data, used by getGuiData()"

    # this creates a dictionary where each key is a FC object name,
    # and each value is a dictionary of property:value pairs,
    # plus a GuiCameraSettings key that contains an iv repr of a coin camera

    def __init__(self):

        super().__init__()
        self.guidata = {}
        self.current = None
        self.properties = {}
        self.currentprop = None
        self.currentval = None
        self.currenttype = None

    # Call when an element starts

    def startElement(self, tag, attributes):

        if tag == "ViewProvider":
            self.current = attributes["name"]
        elif tag == "Property":
            self.currenttype = attributes["type"]
            self.currentprop = attributes["name"]
        elif tag == "Bool":
            if attributes["value"] == "true":
                self.currentval = True
            else:
                self.currentval = False
        elif tag == "PropertyColor":
            c = int(attributes["value"])
            r = float((c >> 24) & 0xFF) / 255.0
            g = float((c >> 16) & 0xFF) / 255.0
            b = float((c >> 8) & 0xFF) / 255.0
            self.currentval = (r, g, b)
        elif tag == "Integer":
            self.currentval = int(attributes["value"])
        elif tag == "String":
            self.currentval = attributes["value"]
        elif tag == "Float":
            self.currentval = float(attributes["value"])
        elif tag == "ColorList":
            self.currentval = attributes["file"]
        elif tag == "PropertyMaterial":
            a = int(attributes["ambientColor"])
            d = int(attributes["diffuseColor"])
            s = int(attributes["specularColor"])
            e = int(attributes["emissiveColor"])
            i = float(attributes["shininess"])
            t = float(attributes["transparency"])
            self.currentval = (a, d, s, e, i, t)
        elif tag == "Enum":
            if isinstance(self.currentval, int):
                self.currentval = [self.currentval, attributes["value"]]
            elif isinstance(self.currentval, list):
                self.currentval.append(attributes["value"])
        elif tag == "Python":
            if "module" in attributes:
                self.currentval = (
                    attributes["value"],
                    attributes["encoded"],
                    attributes["module"],
                    attributes["class"],
                )
        elif tag == "Camera":
            self.guidata["GuiCameraSettings"] = attributes["settings"]

    # Call when an elements ends

    def endElement(self, tag):

        if tag == "ViewProvider":
            if self.current and self.properties:
                self.guidata[self.current] = self.properties
                self.current = None
                self.properties = {}
        elif tag == "Property":
            if self.currentprop and (self.currentval is not None):
                self.properties[self.currentprop] = {
                    "type": self.currenttype,
                    "value": self.currentval,
                }
                self.currentprop = None
                self.currentval = None
                self.currenttype = None


def getGuiData(filename):
    """getGuiData(filename): Extract visual data from a saved FreeCAD file.
    Returns a dictionary ["objectName:dict] where dict contains properties
    keys  like ShapeAppearaance, Transparency, DiffuseColor or Visibility. If found,
    also contains a GuiCameraSettings key with an iv repr of a coin camera"""

    guidata = {}
    zdoc = zipfile.ZipFile(filename)
    if zdoc:
        if "GuiDocument.xml" in zdoc.namelist():
            gf = zdoc.open("GuiDocument.xml")
            guidata = gf.read()
            gf.close()
            Handler = FreeCADGuiHandler()
            xml.sax.parseString(guidata, Handler)
            guidata = Handler.guidata
            for key, properties in guidata.items():
                # open each diffusecolor files and retrieve values
                # first 4 bytes are the array length, then each group of 4 bytes is abgr
                # https://forum.freecad.org/viewtopic.php?t=29382
                if isinstance(properties, dict):
                    for propname in properties.keys():
                        if properties[propname]["type"] == "App::PropertyColorList":
                            if not guidata[key][propname]["value"]:
                                continue
                            df = zdoc.open(guidata[key][propname]["value"])
                            buf = df.read()
                            df.close()
                            cols = []
                            for i in range(1, int(len(buf) / 4)):
                                cols.append(
                                    (
                                        buf[i * 4 + 3] / 255.0,
                                        buf[i * 4 + 2] / 255.0,
                                        buf[i * 4 + 1] / 255.0,
                                        buf[i * 4] / 255.0,
                                    )
                                )
                            guidata[key][propname]["value"] = cols
        zdoc.close()
        # print ("guidata:",guidata)
    return guidata


def saveDiffuseColor(colorlist):
    """saveDiffuseColor(colorlist): Saves the given list or tuple of
    color tuples to a temp file, suitable to include in a DiffuseColor
    property. Returns the path to the created temp file"""

    def tochr(i):
        return bytes((i,))

    # if too many colors, bail out and use only the first one for now...
    if len(colorlist) > 254:
        colorlist = colorlist[:1]
        print("debug: too many colors, reducing")
    output = tochr(len(colorlist)) + 3 * tochr(0)
    allfloats = True
    for color in colorlist:
        for d in color:
            if d > 1:
                allfloats = False
                break
    for color in colorlist:
        if len(color) < 4:
            output += tochr(0)
        for d in reversed(color):
            if allfloats:
                output += tochr(int(d * 255))
            else:
                output += tochr(int(d))
    colfile = tempfile.mkstemp(prefix="DiffuseColor")[-1]
    f = open(colfile, "wb")
    f.write(output)
    f.close()
    return colfile


def getColors(filename, nodiffuse=False):
    """getColors(filename,nodiffuse): Extracts the colors saved in a FreeCAD file
    Returns a dictionary containing ["objectName":colors] pairs.
    colrs can be either a 3-element tuple representing an RGB color, if
    the object has no per-face colors (DiffuseColor) defined, or a list
    of tuples if per-face colors are available. In case of DiffuseColors,
    tuples can have 4 values (RGBT) (T = transparency, inverse of alpha)
    This is a reduced version of getGuiData(), which returns more information.
    If nodiffuse = True, DiffuseColor info is discarded, only ShapeColor is read."""

    guidata = getGuiData(filename)
    colors = {}
    for k, v in guidata.items():
        if ("DiffuseColor" in v) and (not nodiffuse):
            if len(v["DiffuseColor"]["value"]) == 1:
                # only one color in DiffuseColor: used for the whole object
                colors[k] = v["DiffuseColor"]["value"][0]
            else:
                colors[k] = v["DiffuseColor"]["value"]
        elif "ShapeColor" in v:
            colors[k] = v["ShapeColor"]["value"]
    return colors


def getStepData(objects, colors):
    """getStepData(objects,colors): transforms the given list of objects and
    colors dictionary into a list of tuples acceptable by the STEP exporter of
    FreeCAD's Import module"""

    # The STEP exporter accepts two kinds of data: [obj1,obj2,...] list, or
    # [(obj1,DiffuseColor),(obj2,DiffuseColor),...] list.

    # we need to reformat a bit the data we send to the exporter
    # since, differently than the others, it wants (object,DiffuseColor) tuples
    data = []
    for obj in objects:
        if obj.Name in colors:
            color = colors[obj.Name]
            if isinstance(color, tuple):
                # this is a ShapeAppeaaraance. We reformat as a list so it works as a DiffuseColor,
                # which is what the exporter expects. DiffuseColor can have either one color,
                # or the same number of colors as the number of faces
                color = [color]
            data.append((obj, color))
        else:
            # no color information. This object will be exported without colors
            data.append(obj)
    return data


def render(
    outputfile,
    scene=None,
    camera=None,
    zoom=False,
    width=400,
    height=300,
    background=(1.0, 1.0, 1.0),
    lightdir=None,
):
    """render(outputfile,scene=None,camera=None,zoom=False,width=400,height=300,background=(1.0,1.0,1.0),lightdir=None):
    Renders a PNG image of given width and height and background color from the given coin scene, using
    the given coin camera (ortho or perspective). If zoom is True the camera will be resized to fit all
    objects. The outputfile must be a file path to save a png image. Optionally a light direction as a (x,y,z)
    tuple can be given. In this case, a directional light will be added and shadows will
    be turned on. This might not work with some 3D drivers."""

    # On Linux, the X server must have indirect rendering enabled in order to be able to do offline
    # PNG rendering. Unfortunately, this is turned off by default on most recent distros. The easiest
    # way I found is to edit (or create if inexistent) /etc/X11/xorg.conf and add this:
    #
    # Section "ServerFlags"
    #    Option "AllowIndirectGLX" "on"
    #    Option "IndirectGLX" "on"
    # EndSection
    #
    # But there are other ways, google of GLX indirect rendering

    from pivy import coin

    if isinstance(camera, str):
        camera = getCoinCamera(camera)

    print("Starting offline renderer")
    # build an offline scene root separator
    root = coin.SoSeparator()
    if not lightdir:
        # add one light (mandatory)
        light = coin.SoDirectionalLight()
        root.addChild(light)
    if not camera:
        # create a default camera if none was given
        camera = coin.SoPerspectiveCamera()
        cameraRotation = coin.SbRotation.identity()
        cameraRotation *= coin.SbRotation(coin.SbVec3f(1, 0, 0), 1.0)
        cameraRotation *= coin.SbRotation(coin.SbVec3f(0, 0, 1), 0.4)
        camera.orientation = cameraRotation
        # make sure all objects get in the view later
        zoom = True
    root.addChild(camera)
    if scene:
        root.addChild(scene)
    else:
        # no scene was given, add a simple cube
        cube = coin.SoCube()
        root.addChild(cube)
    if lightdir:
        root = embedLight(root, lightdir)
    vpRegion = coin.SbViewportRegion(width, height)
    if zoom:
        camera.viewAll(root, vpRegion)
    print("Creating viewport")
    offscreenRenderer = coin.SoOffscreenRenderer(vpRegion)
    offscreenRenderer.setBackgroundColor(coin.SbColor(background[0], background[1], background[2]))
    print("Ready to render")
    # ref ensures that the node will not be garbage-collected during rendering
    root.ref()
    ok = offscreenRenderer.render(root)
    root.unref()
    if ok:
        offscreenRenderer.writeToFile(outputfile, "PNG")
        print("Rendering", outputfile, "done")
    else:
        print("Error rendering image")


def buildScene(objects, colors=None):
    """buildScene(objects,colors=None): builds a coin node from a given list of FreeCAD
    objects. Optional colors argument can be a dictionary of objName:ShapeColorTuple
    or obj:DiffuseColorList pairs."""

    from pivy import coin

    root = coin.SoSeparator()
    for o in objects:
        buf = None
        if hasattr(o, "Shape") and o.Shape and (not o.Shape.isNull()):
            # writeInventor of shapes needs tessellation values
            buf = o.Shape.writeInventor(2, 0.01)
        elif o.isDerivedFrom("Mesh::Feature"):
            buf = o.Mesh.writeInventor()
        if buf:
            # 3 lines below are the standard way to produce a coin node from a string
            # Part and Mesh objects always have a containing SoSeparator
            inp = coin.SoInput()
            inp.setBuffer(buf)
            node = coin.SoDB.readAll(inp)
            if colors:
                if o.Name in colors:
                    # insert a material node at 1st position, before the geometry
                    color = colors[o.Name]
                    if isinstance(color, list):
                        # DiffuseColor, not supported here
                        color = color[0]
                    color = color[:3]
                    mat = coin.SoMaterial()
                    mat.diffuseColor = color
                    node.insertChild(mat, 0)
            root.addChild(node)
    return root


def getCamera(filepath):
    """getCamera(filepath): Returns a string representing a coin camera node from a given FreeCAD
    file, or None if none was found inside"""

    guidata = getGuiData(filepath)
    if "GuiCameraSettings" in guidata:
        return guidata["GuiCameraSettings"].strip()
    print("No camera found in file")
    return None


def getCoinCamera(camerastring):
    """getCoinCamera(camerastring): Returns a coin camera node from a string"""

    from pivy import coin

    if camerastring:
        inp = coin.SoInput()
        inp.setBuffer(camerastring)
        node = coin.SoDB.readAll(inp)
        # this produces a SoSeparator containing the camera
        for child in node.getChildren():
            if ("SoOrthographicCamera" in str(child)) or ("SoPerspectiveCamera" in str(child)):
                return child
    print("unable to build a camera node from string:", camerastring)
    return None


def viewer(scene=None, background=(1.0, 1.0, 1.0), lightdir=None):
    """viewer(scene=None,background=(1.0,1.0,1.0),lightdir=None): starts
    a standalone coin viewer with the contents of the given scene. You can
    give a background color, and optionally a light direction as a (x,y,z)
    tuple. In this case, a directional light will be added and shadows will
    be turned on. This might not work with some 3D drivers."""

    # Initialize Coin. This returns a main window to use
    from pivy import coin
    from pivy import quarter
    from PySide2 import QtWidgets

    app = QtWidgets.QApplication([])
    # Create a Qt widget, which will be our window.
    win = quarter.QuarterWidget()

    if win is None:
        print("Unable to create a Quarter window")
        return

    win.setBackgroundColor(coin.SbColor(background[0], background[1], background[2]))

    if not scene:
        # build a quick default scene
        mat = coin.SoMaterial()
        mat.diffuseColor = (1.0, 0.0, 0.0)
        # Make a scene containing a red cone
        scene = coin.SoSeparator()
        scene.addChild(mat)
        scene.addChild(coin.SoCone())

    if lightdir:
        scene = embedLight(scene, lightdir)

    # ref the scene so it doesn't get garbage-collected
    scene.ref()

    # Put our scene into viewer, change the title
    win.setSceneGraph(scene)
    win.setWindowTitle("Coin viewer")

    # IMPORTANT!!!!! Windows are hidden by default.
    win.show()
    # Start the event loop.
    app.exec_()


def embedLight(scene, lightdir):
    """embedLight(scene,lightdir): embeds a given coin node
    inside a shadow group with directional light with the
    given direction (x,y,z) tuple. Returns the final coin node"""

    from pivy import coin

    # buggy - no SoShadowGroup in pivy?
    # sgroup = coin.SoShadowGroup()
    # sgroup.quality = 1
    # sgroup.precision = 1
    # slight = SoShadowDirectionalLight()
    # slight.direction = lightdir
    # slight.intensity = 20.0

    buf = (
        """
#Inventor V2.1 ascii
ShadowGroup {
    quality 1
    precision 1
    ShadowDirectionalLight {
        direction """
        + str(lightdir[0])
        + " "
        + str(lightdir[1])
        + " "
        + str(lightdir[2])
        + """
        intensity 200.0
        # enable this to reduce the shadow view distance
        # maxShadowDistance 200
    }
}"""
    )

    inp = coin.SoInput()
    inp.setBuffer(buf)
    sgroup = coin.SoDB.readAll(inp)

    sgroup.addChild(scene)
    return sgroup


def save(document, filename=None, guidata=None, colors=None, camera=None):
    """save(document,filename=None,guidata=None,colors=None,camera=None): Saves the current document. If no filename
    is given, the filename stored in the document (document.FileName) is used.

    You can provide a guidata dictionary, which can be obtained by the getGuiData() function, and has the form:

     { "objectName" :
         { "propertyName" :
             { "type"  : "App::PropertyString",
               "value" : "My Value"
             }
         }
     }

    The type of the "value" contents depends on the type (int, string, float,tuple...) see inside the FreeCADGuiHandler
    class to get an idea.

    If guidata is provided, colors and camera attributes are discarded.

    Alternatively, a color dictionary of objName:ShapeColorTuple or obj:DiffuseColorList pairs.can be provided,
    in that case the objects will keep their colors when opened in the FreeCAD GUI. If given, camera is a string
    representing a coin camera node."""

    if filename:
        print("Saving as", filename)
        document.saveAs(filename)
    else:
        if document.FileName:
            filename = document.FileName
            document.save()
        else:
            print("Unable to save this document. Provide a file name")
            return

    if guidata:
        guidocs = buildGuiDocumentFromGuiData(document, guidata)
        if guidocs:
            zf = zipfile.ZipFile(filename, mode="a")
            for guidoc in guidocs:
                zf.write(guidoc[0], guidoc[1])
            zf.close()
            # delete the temp files
            for guidoc in guidocs:
                os.remove(guidoc[0])
    elif colors:
        guidoc = buildGuiDocumentFromColors(document, colors, camera)
        if guidoc:
            zf = zipfile.ZipFile(filename, mode="a")
            zf.write(guidoc[0], "GuiDocument.xml")
            for colorfile in guidoc[1:]:
                zf.write(colorfile, os.path.basename(colorfile))
            zf.close()
            # delete the temp files
            for tfile in guidoc:
                os.remove(tfile)


def getUnsigned(color):
    """getUnsigned(color): returns an unsigned int from a (r,g,b) color tuple"""

    if (color[0] <= 1) and (color[1] <= 1) and (color[2] <= 1):
        # 0->1 float colors, convert to 0->255
        color = (color[0] * 255.0, color[1] * 255.0, color[2] * 255.0)

    # ensure everything is int otherwise bit ops below don't work
    color = (int(color[0]), int(color[1]), int(color[2]))

    # https://forum.freecad.org/viewtopic.php?t=19074
    return str(color[0] << 24 | color[1] << 16 | color[2] << 8)


def buildGuiDocumentFromColors(document, colors, camera=None):
    """buildGuiDocumentFromColors(document,colors,camera=None): Returns the path to a temporary GuiDocument.xml for the given document.
    Colors is a color dictionary of objName:ShapeColorTuple or obj:DiffuseColorList. Camera, if given, is a string representing
    a coin camera. You must delete the temporary file after using it."""

    if not camera:
        camera = "OrthographicCamera {   viewportMapping ADJUST_CAMERA   position 0 -0 20000   orientation 0.0, 0.8939966636005564, 0.0, -0.44807361612917324   nearDistance 7561.228   farDistance 63175.688   aspectRatio 1   focalDistance 35368.102   height 2883.365  }"

    guidoc = "<?xml version='1.0' encoding='utf-8'?>\n"
    guidoc += "<!--\n"
    guidoc += " FreeCAD Document, see https://www.freecad.org for more information…\n"
    guidoc += "-->\n"
    guidoc += '<Document SchemaVersion="1">\n'

    colfiles = []

    vps = [obj for obj in document.Objects if obj.Name in colors]
    if not vps:
        return None
    guidoc += '    <ViewProviderData Count="' + str(len(vps)) + '">\n'
    for vp in vps:
        guidoc += '        <ViewProvider name="' + vp.Name + '" expanded="0">\n'
        guidoc += '            <Properties Count="2">\n'
        vpcol = colors[vp.Name]
        if isinstance(vpcol[0], tuple):
            # distinct diffuse colors
            colfile = saveDiffuseColor(vpcol)
            name = os.path.basename(colfile)
            colfiles.append(colfile)
            guidoc += (
                '                <Property name="DiffuseColor" type="App::PropertyColorList">\n'
            )
            guidoc += '                    <ColorList file="' + name + '"/>\n'
            guidoc += "                </Property>\n"
        else:
            guidoc += '                <Property name="ShapeColor" type="App::PropertyColor">\n'
            guidoc += '                    <PropertyColor value="' + getUnsigned(vpcol) + '"/>\n'
            guidoc += "                </Property>\n"
        guidoc += '                <Property name="Visibility" type="App::PropertyBool">\n'
        guidoc += '                    <Bool value="true"/>\n'
        guidoc += "                </Property>\n"
        if hasattr(vp, "Proxy"):
            # if this is a python feature, store the view provider class if possible
            m = getViewProviderClass(vp)
            if m:
                guidoc += (
                    '                <Property name="Proxy" type="App::PropertyPythonObject">\n'
                )
                guidoc += (
                    '                    <Python value="bnVsbA==" encoded="yes" module="'
                    + m[0]
                    + '" class="'
                    + m[1]
                    + '"/>\n'
                )
                guidoc += "                </Property>\n"
        guidoc += "            </Properties>\n"
        guidoc += "        </ViewProvider>\n"

    guidoc += "    </ViewProviderData>\n"
    guidoc += '    <Camera settings="  ' + camera + ' "/>\n'
    guidoc += "</Document>"

    # although the zipfile module has a writestr() function that should allow us to write the
    # string above directly to the zip file, I couldn't manage to make it work.. So we rather
    # use a temp file here, which works.

    # print(guidoc)

    tempxml = tempfile.mkstemp(suffix=".xml")[-1]
    f = open(tempxml, "w")
    f.write(guidoc)
    f.close()

    return [tempxml] + colfiles


def buildGuiDocumentFromGuiData(document, guidata):
    """buildGuiDocumentFromColors(document,guidata): Returns the path to a temporary GuiDocument.xml for the given document.

       GuiData is a dictionary, which can be obtained by the getGuiData() function, and has the form:

        { "objectName" :
            { "propertyName" :
                { "type"  : "App::PropertyString",
                  "value" : "My Value"
                }
            }
        }
    This function returns a list of (filepath,name) tuples, the first one named GuiDocument.xml and the next ones being color files
    """

    files = []
    colorindex = 1

    guidoc = "<?xml version='1.0' encoding='utf-8'?>\n"
    guidoc += "<!--\n"
    guidoc += " FreeCAD Document, see https://www.freecad.org for more information…\n"
    guidoc += "-->\n"
    guidoc += '<Document SchemaVersion="1">\n'

    vps = [obj for obj in document.Objects if obj.Name in guidata]
    if not vps:
        return None
    guidoc += '    <ViewProviderData Count="' + str(len(vps)) + '">\n'
    for vp in vps:
        properties = guidata[vp.Name]
        guidoc += '        <ViewProvider name="' + vp.Name + '" expanded="0">\n'
        guidoc += '            <Properties Count="' + str(len(properties)) + '">\n'
        for name, prop in properties.items():
            guidoc += '                <Property name="' + name + '" type="' + prop["type"] + '">\n'
            if prop["type"] in ["App::PropertyString", "PropertyFont"]:
                guidoc += '                    <String value="' + prop["value"] + '"/>\n'
            elif prop["type"] in [
                "App::PropertyAngle",
                "App::PropertyFloat",
                "App::PropertyFloatConstraint",
                "App::PropertyDistance",
                "App::PropertyLength",
            ]:
                guidoc += '                    <Float value="' + str(prop["value"]) + '"/>\n'
            elif prop["type"] in ["App::PropertyInteger", "App::PropertyPercent"]:
                guidoc += '                    <Integer value="' + str(prop["value"]) + '"/>\n'
            elif prop["type"] in ["App::PropertyBool"]:
                guidoc += '                    <Bool value="' + str(prop["value"]).lower() + '"/>\n'
            elif prop["type"] in ["App::PropertyEnumeration"]:
                if isinstance(prop["value"], int):
                    guidoc += '                    <Integer value="' + str(prop["value"]) + '"/>\n'
                elif isinstance(prop["value"], list):
                    guidoc += (
                        '                    <Integer value="'
                        + str(prop["value"][0])
                        + '" CustomEnum="true"/>\n'
                    )
                    guidoc += (
                        '                    <CustomEnumList count="'
                        + str(len(prop["value"]) - 1)
                        + '">\n'
                    )
                    for v in prop["value"][1:]:
                        guidoc += '                        <Enum value="' + v + '"/>\n'
                    guidoc += "                    </CustomEnumList>\n"
            elif prop["type"] in ["App::PropertyColorList"]:
                # DiffuseColor: first 4 bytes of file tells number of Colors
                # then rest of bytes represent colors value where each color
                # is of 4 bytes in abgr order.

                # convert number of colors into hexadecimal representation
                hex_repr = hex(len(prop["value"]))[2:]

                # if len of `hex_repr` is odd, then add 0 padding.
                # `hex_repr` must contain an even number of hex digits
                # as `binascii.unhexlify` only works of even hex str.
                if len(hex_repr) % 2 != 0:
                    hex_repr = "0" + hex_repr
                buf = binascii.unhexlify(hex_repr)

                if len(hex_repr) > 8:
                    raise NotImplementedError(
                        "Number of colors ({}) is greater than 4 bytes and in DiffuseColor file we only "
                        "specify number of colors in 4 bytes.".format(len(prop["value"]))
                    )
                elif len(hex_repr) == 2:  # `hex_repr` == 1 byte
                    # fill 3 other bytes (the number of colors occupies 4 bytes)
                    buf += binascii.unhexlify(hex(0)[2:].zfill(2))
                    buf += binascii.unhexlify(hex(0)[2:].zfill(2))
                    buf += binascii.unhexlify(hex(0)[2:].zfill(2))
                elif len(hex_repr) == 4:  # `hex_repr` == 2 bytes
                    buf += binascii.unhexlify(hex(0)[2:].zfill(2))
                    buf += binascii.unhexlify(hex(0)[2:].zfill(2))
                elif len(hex_repr) == 6:  # `hex_repr` == 3 bytes
                    buf += binascii.unhexlify(hex(0)[2:].zfill(2))

                # fill colors in abgr order
                for color in prop["value"]:
                    if len(color) >= 4:
                        buf += binascii.unhexlify(hex(int(color[3] * 255))[2:].zfill(2))
                    else:
                        buf += binascii.unhexlify(hex(0)[2:].zfill(2))
                    buf += binascii.unhexlify(hex(int(color[2] * 255))[2:].zfill(2))
                    buf += binascii.unhexlify(hex(int(color[1] * 255))[2:].zfill(2))
                    buf += binascii.unhexlify(hex(int(color[0] * 255))[2:].zfill(2))
                tempcolorfile = tempfile.mkstemp(suffix=".xml")[-1]
                f = open(tempcolorfile, "wb")
                f.write(buf)
                f.close()
                tempcolorname = "ColorFile" + str(colorindex)
                colorindex += 1
                guidoc += '                    <ColorList file="' + tempcolorname + '"/>\n'
                files.append((tempcolorfile, tempcolorname))
            elif prop["type"] in ["App::PropertyMaterial"]:
                guidoc += '                    <PropertyMaterial ambientColor="' + str(
                    prop["value"][0]
                )
                guidoc += (
                    '" diffuseColor="'
                    + str(prop["value"][1])
                    + '" specularColor="'
                    + str(prop["value"][2])
                )
                guidoc += (
                    '" emissiveColor="'
                    + str(prop["value"][3])
                    + '" shininess="'
                    + str(prop["value"][4])
                )
                guidoc += '" transparency="' + str(prop["value"][5]) + '"/>\n'
            elif prop["type"] in ["App::PropertyPythonObject"]:
                guidoc += (
                    '                    <Python value="' + str(prop["value"][0]) + '" encoded="'
                )
                guidoc += str(prop["value"][1]) + '" module="' + str(prop["value"][2]) + '" class="'
                guidoc += str(prop["value"][3]) + '"/>\n'
            elif prop["type"] in ["App::PropertyColor"]:
                guidoc += (
                    '                    <PropertyColor value="'
                    + str(getUnsigned(prop["value"]))
                    + '"/>\n'
                )
            guidoc += "                </Property>\n"
        guidoc += "            </Properties>\n"
        guidoc += "        </ViewProvider>\n"
    guidoc += "    </ViewProviderData>\n"
    if "GuiCameraSettings" in guidata:
        guidoc += '    <Camera settings="  ' + guidata["GuiCameraSettings"] + ' "/>\n'
    guidoc += "</Document>\n"

    # although the zipfile module has a writestr() function that should allow us to write the
    # string above directly to the zip file, I couldn't manage to make it work.. So we rather
    # use a temp file here, which works.

    # print(guidoc)

    tempxml = tempfile.mkstemp(suffix=".xml")[-1]
    f = open(tempxml, "w")
    f.write(guidoc)
    f.close()
    files.insert(0, (tempxml, "GuiDocument.xml"))
    return files


def getViewProviderClass(obj):
    """getViewProviderClass(obj): tries to identify the associated view provider for a
    given python object. Returns a (modulename,classname) tuple if found, or None"""

    if not hasattr(obj, "Proxy"):
        return None
    if not obj.Proxy:
        return None
    mod = obj.Proxy.__module__
    objclass = obj.Proxy.__class__.__name__
    classes = []
    for name, mem in inspect.getmembers(sys.modules[mod]):
        if inspect.isclass(mem):
            classes.append(mem.__name__)
    # try to find a matching ViewProvider class
    if objclass.startswith("_"):
        wantedname = "_ViewProvider" + objclass[1:]
    else:
        wantedname = "ViewProvider" + objclass
    if wantedname in classes:
        # print("Found matching view provider for",mod,objclass,wantedname)
        return (
            mod,
            wantedname,
        )
    # use the default Draft VP if this is a Draft object
    if mod == "Draft":
        return (mod, "_ViewProviderDraft")
    print("Found no matching view provider for", mod, objclass)
    return None


def extract(filename, inputpath, outputpath=None):
    """extract(filename,inputpath,outputpath=None): extracts 'inputpath' which is a filename
    stored in filename (a FreeCAD or zip file). If outputpath is given, the file is saved as outputpath and
    nothing is returned. If not, the contents of the inputfile are returned and nothing is saved."""

    zdoc = zipfile.ZipFile(filename)
    if zdoc:
        if inputpath in zdoc.namelist():
            gf = zdoc.open(inputpath)
            data = gf.read()
            gf.close()
            if data:
                if outputpath:
                    if isinstance(data, str):
                        of = open(outputpath, "w")
                    else:
                        of = open(outputpath, "wb")
                    of.write(data)
                    of.close()
                else:
                    return data


def openiv(filename):
    """openiv(filename): opens an .iv file and returns a coin node from it"""

    from pivy import coin

    f = open(filename, "r")
    buf = f.read()
    f.close()
    inp = coin.SoInput()
    inp.setBuffer(buf)
    node = coin.SoDB.readAll(inp)
    return node


def saveiv(scene, filename):
    """saveiv(scene,filename): saves an .iv file with the contents of the given coin node"""

    from pivy import coin

    wa = coin.SoWriteAction()
    wa.getOutput().openFile(filename)
    wa.getOutput().setBinary(False)
    wa.apply(scene)
    wa.getOutput().closeFile()


##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2024 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to scale objects in the 3D space.

The scale operation can also be done with subelements.

The subelements operations only really work with polylines (Wires)
because internally the functions `scale_vertex` and `scale_edge`
only work with polylines that have a `Points` property.
"""
## @package gui_scale
# \ingroup draftguitools
# \brief Provides GUI tools to scale objects in the 3D space.

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
import DraftVecUtils
from draftguitools import gui_base_original
from draftguitools import gui_tool_utils
from draftguitools import gui_trackers as trackers
from draftutils import groups
from draftutils import params
from draftutils import utils
from draftutils import todo
from draftutils.messages import _msg, _err, _toolmsg
from draftutils.translate import translate
from drafttaskpanels import task_scale


class Scale(gui_base_original.Modifier):
    """Gui Command for the Scale tool.

    This tool scales the selected objects from a base point.
    """

    def GetResources(self):
        """Set icon, menu and tooltip."""
        return {
            "Pixmap": "Draft_Scale",
            "Accel": "S, C",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Scale", "Scale"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Scale", "Scales the selected objects from a base point"
            ),
        }

    def Activated(self):
        """Execute when the command is called."""
        super().Activated(name="Scale")
        if not self.ui:
            return
        self.ghosts = []
        self.get_object_selection()

    def get_object_selection(self):
        """Get object selection and proceed if successful."""
        if Gui.Selection.getSelection():
            return self.proceed()
        self.ui.selectUi(on_close_call=self.finish)
        _msg(translate("draft", "Select an object to scale"))
        self.call = self.view.addEventCallback("SoEvent", gui_tool_utils.selectObject)

    def proceed(self):
        """Proceed with execution of the command after selection."""
        if self.call:
            self.view.removeEventCallback("SoEvent", self.call)
        self.selection = Gui.Selection.getSelectionEx("", 0)
        Gui.doCommand('selection = FreeCADGui.Selection.getSelectionEx("", 0)')
        self.refs = []
        self.ui.pointUi(title=translate("draft", self.featureName), icon="Draft_Scale")
        self.ui.isRelative.hide()
        self.ui.xValue.setFocus()
        self.ui.xValue.selectAll()
        self.pickmode = False
        self.task = None
        self.call = self.view.addEventCallback("SoEvent", self.action)
        _toolmsg(translate("draft", "Pick base point"))

    def set_ghosts(self):
        """Set the ghost to display."""
        for ghost in self.ghosts:
            ghost.remove()
        if self.task is None:
            copy = params.get_param("ScaleCopy")
            clone = params.get_param("ScaleClone")
            subelements = params.get_param("SubelementMode")
        else:
            copy = self.task.isCopy.isChecked()
            clone = self.task.isClone.isChecked()
            subelements = self.task.isSubelementMode.isChecked()
        if subelements:
            self.ghosts = self.get_subelement_ghosts(self.selection, copy)
            if not self.ghosts:
                _err(translate("draft", "No valid subelements selected"))
        else:
            objs, places, _ = utils._modifiers_process_selection(
                self.selection, (copy or clone), scale=True
            )
            self.ghosts = [trackers.ghostTracker(objs, parent_places=places)]

    def get_subelement_ghosts(self, selection, copy):
        """Get ghost for the subelements (vertices, edges)."""
        import Part

        ghosts = []
        for sel in selection:
            for sub in sel.SubElementNames if sel.SubElementNames else [""]:
                if (not copy and "Vertex" in sub) or "Edge" in sub:
                    shape = Part.getShape(sel.Object, sub, needSubElement=True, retType=0)
                    ghosts.append(trackers.ghostTracker(shape))
        return ghosts

    def scale_ghosts(self, x, y, z, rel=False):
        """Scale the preview of the object."""
        delta = App.Vector(x, y, z)
        # ScaleRelative option removed in v1.1 as it does not work properly:
        # if rel:
        # delta = self.wp.get_local_coords(delta)
        for ghost in self.ghosts:
            ghost.scale(delta)
        # calculate a correction factor depending on the scaling center
        corr = App.Vector(self.node[0])
        corr.scale(*delta)
        corr = (corr.sub(self.node[0])).negative()
        for ghost in self.ghosts:
            ghost.flip_normals(x * y * z < 0)
            ghost.move(corr)
            ghost.on()

    def pick_ref(self):
        """Pick a point of reference."""
        self.pickmode = True
        if self.node:
            self.node = self.node[:1]  # remove previous picks
        _toolmsg(translate("draft", "Pick reference distance from base point"))
        self.call = self.view.addEventCallback("SoEvent", self.action)

    def action(self, arg):
        """Handle the 3D scene events.

        This is installed as an EventCallback in the Inventor view.

        Parameters
        ----------
        arg: dict
            Dictionary with strings that indicates the type of event received
            from the 3D view.
        """
        if arg["Type"] == "SoKeyboardEvent" and arg["Key"] == "ESCAPE":
            self.finish()
        elif arg["Type"] == "SoLocation2Event":
            self.handle_mouse_move_event(arg)
        elif (
            arg["Type"] == "SoMouseButtonEvent"
            and arg["State"] == "DOWN"
            and arg["Button"] == "BUTTON1"
            and self.point
        ):
            self.handle_mouse_click_event()

    def handle_mouse_move_event(self, arg):
        """Handle the mouse event of movement."""
        for ghost in self.ghosts:
            ghost.off()
        self.point, ctrlPoint, info = gui_tool_utils.getPoint(self, arg)

    def handle_mouse_click_event(self):
        """Handle the mouse click event."""
        if not self.ghosts:
            self.set_ghosts()
        self.numericInput(self.point.x, self.point.y, self.point.z)

    def scale(self):
        """Perform the scale of the object.

        Scales the subelements, or with a clone, or just general scaling.
        """
        sx = self.task.xValue.value()
        sy = self.task.yValue.value()
        sz = self.task.zValue.value()
        if sx * sy * sz == 0:
            _err(translate("draft", "Zero scale factor not allowed"))
            self.finish()
            return

        self.delta = App.Vector(sx, sy, sz)
        # ScaleRelative option removed in v1.1 as it does not work properly:
        # if self.task.relative.isChecked():
        # self.delta = self.wp.get_local_coords(self.delta)
        self.center = self.node[0]
        if self.task.isCopy.isChecked():
            cmd_name = translate("draft", "Copy")
        else:
            cmd_name = translate("draft", "Scale")
        Gui.addModule("Draft")
        cmd = "Draft.scale(selection, "
        cmd += "scale=" + DraftVecUtils.toString(self.delta) + ", "
        cmd += "center=" + DraftVecUtils.toString(self.center) + ", "
        cmd += "copy=" + str(self.task.isCopy.isChecked()) + ", "
        cmd += "clone=" + str(self.task.isClone.isChecked()) + ", "
        cmd += "subelements=" + str(self.task.isSubelementMode.isChecked()) + ")"
        cmd_list = [cmd, "FreeCAD.ActiveDocument.recompute()"]
        self.commit(cmd_name, cmd_list)
        self.finish()

    def numericInput(self, numx, numy, numz):
        """Validate the entry fields in the user interface.

        This function is called by the toolbar or taskpanel interface
        when valid x, y, and z have been entered in the input fields.
        """
        self.point = App.Vector(numx, numy, numz)
        self.node.append(self.point)
        if not self.pickmode:
            if not self.ghosts:
                self.set_ghosts()
            self.ui.offUi()
            if self.call:
                self.view.removeEventCallback("SoEvent", self.call)
            self.task = task_scale.ScaleTaskPanel()
            self.task.sourceCmd = self
            todo.ToDo.delay(Gui.Control.showDialog, self.task)
            todo.ToDo.delay(self.task.xValue.selectAll, None)
            todo.ToDo.delay(self.task.xValue.setFocus, None)
            for ghost in self.ghosts:
                ghost.on()
        elif len(self.node) == 2:
            _toolmsg(translate("draft", "Pick new distance from base point"))
        elif len(self.node) == 3:
            if hasattr(Gui, "Snapper"):
                Gui.Snapper.off()
            if self.call:
                self.view.removeEventCallback("SoEvent", self.call)
            d1 = (self.node[1].sub(self.node[0])).Length
            d2 = (self.node[2].sub(self.node[0])).Length
            # print ("d2/d1 = {}".format(d2/d1))
            if hasattr(self, "task"):
                if self.task:
                    self.task.lock.setChecked(True)
                    self.task.setValue(d2 / d1)

    def finish(self, cont=False):
        """Terminate the operation."""
        self.end_callbacks(self.call)
        for ghost in self.ghosts:
            ghost.finalize()
        super().finish()


Gui.addCommand("Draft_Scale", Scale())

## @}



# ***************************************************************************
# *   Copyright (c) 2017 Markus Hovorka <m.hovorka@live.de>                 *
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *   Copyright (c) 2022 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM solver Elmer writer"
__author__ = "Markus Hovorka, Bernd Hahnebach, Uwe Stöhr"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

import os
import os.path
import subprocess
import tempfile
from platform import system

from FreeCAD import Console
from FreeCAD import Units
from FreeCAD import ParamGet

import Fem
from . import sifio
from .. import settings
from femtools import constants
from femtools import femutils
from femtools import membertools
from .equations import deformation_writer as DEF_writer
from .equations import elasticity_writer as EL_writer
from .equations import electricforce_writer as EF_writer
from .equations import electrostatic_writer as ES_writer
from .equations import flow_writer
from .equations import flux_writer
from .equations import heat_writer
from .equations import magnetodynamic_writer as MgDyn_writer
from .equations import magnetodynamic2D_writer as MgDyn2D_writer
from .equations import staticcurrent_writer as SC_writer


_STARTINFO_NAME = "ELMERSOLVER_STARTINFO"
_SIF_NAME = "case.sif"
_ELMERGRID_IFORMAT = "8"
_ELMERGRID_OFORMAT = "2"
_COORDS_NON_MAGNETO_2D = [
    "Polar 2D",
    "Polar 3D",
    "Cartesian 3D",
    "Cylindric",
    "Cylindric Symmetric",
]


def _getAllSubObjects(obj):
    s = ["Solid" + str(i + 1) for i in range(len(obj.Shape.Solids))]
    s.extend("Face" + str(i + 1) for i in range(len(obj.Shape.Faces)))
    s.extend("Edge" + str(i + 1) for i in range(len(obj.Shape.Edges)))
    s.extend("Vertex" + str(i + 1) for i in range(len(obj.Shape.Vertexes)))
    return s


class Writer:

    def __init__(self, solver, directory, testmode=False):
        self.analysis = solver.getParentGroup()
        self.solver = solver
        self.directory = directory
        Console.PrintMessage(f"Write elmer input files to: {self.directory}\n")
        self.testmode = testmode
        self._usedVarNames = set()
        self._builder = sifio.Builder()
        self._handledObjects = set()
        self._handleUnits()
        self._handleConstants()

    def getHandledConstraints(self):
        return self._handledObjects

    def _writeBlocks(self):
        self._handleRedifinedConstants()
        self._handleSimulation()
        self._handleDeformation()
        self._handleElasticity()
        self._handleElectricforce()
        self._handleElectrostatic()
        self._handleHeat()
        self._handleFlow()
        self._handleFlux()
        self._handleMagnetodynamic()
        self._handleMagnetodynamic2D()
        self._handleStaticCurrent()
        self._addOutputSolver()

    def write_solver_input(self):
        self._writeBlocks()
        self._writeSif()
        self._writeStartinfo()

    def _handleUnits(self):
        # Elmer solver writer no longer uses FreeCAD unit system
        # to retrieve units for writing the sif file
        #
        # ATM Elmer writer uses SI units only
        #
        # see forum topic: https://forum.freecad.org/viewtopic.php?f=18&t=70150
        #
        # TODO: adapt method and comment
        # should be only one system for all solver and not in each solver
        # https://forum.freecad.org/viewtopic.php?t=47895
        # https://forum.freecad.org/viewtopic.php?t=48451
        # https://forum.freecad.org/viewtopic.php?f=10&t=48642
        # The FreeCAD unit schema is only used to determine the schema number
        # all definition are done here ATM
        # keep in mind a unit schema might not be consistent:
        # Length could be mm and Area could be m2 and Volume could be cm3
        # as long as only the seven base units are retrieved from a unit schema
        # the units are consistent
        # TODO retrieve the seven base units from FreeCAD unit schema
        # instead of hard coding them here for a second once
        self.unit_schema = Units.Scheme.Internal
        self.unit_system = {  # standard FreeCAD Base units = unit schema 0
            "L": "m",
            "M": "kg",
            "T": "s",
            "I": "A",
            "O": "K",
            "N": "mol",
            "J": "cd",
        }
        param = ParamGet("User parameter:BaseApp/Preferences/Units")
        self.unit_schema = param.GetInt("UserSchema", Units.Scheme.Internal)
        if self.unit_schema == Units.Scheme.Internal:
            Console.PrintMessage(
                "The FreeCAD standard unit schema mm/kg/s is used. "
                "Elmer sif-file writing is however done in SI units.\n"
            )
        elif self.unit_schema == Units.Scheme.MKS:
            Console.PrintMessage(
                "The SI unit schema m/kg/s is used. "
                "Elmer sif-file writing is done in SI-units.\n"
            )
            self.unit_system = {
                "L": "m",
                "M": "kg",
                "T": "s",
                "I": "A",
                "O": "K",
                "N": "mol",
                "J": "cd",
            }
        elif self.unit_schema == Units.Scheme.FEM:
            # see also unit comment in calculix writer
            Console.PrintMessage(
                "The FEM unit schema mm/N/s is used. "
                "Elmer sif-file writing is however done in SI units.\n"
            )
            self.unit_system = {
                "L": "m",
                "M": "kg",
                "T": "s",
                "I": "A",
                "O": "K",
                "N": "mol",
                "J": "cd",
            }
        elif self.unit_schema > Units.Scheme.MKS and self.unit_schema != Units.Scheme.FEM:
            Console.PrintMessage(
                "Unit schema: {} not supported by Elmer writer. "
                "The FreeCAD standard unit schema mm/kg/s is used. "
                "Elmer sif-file writing is done in Standard FreeCAD units.\n".format(
                    Units.listSchemas(self.unit_schema)
                )
            )

    def getFromUi(self, value, unit, outputDim):
        quantity = Units.Quantity(str(value) + str(unit))
        return self.convert(quantity, outputDim)

    def convert(self, quantityStr, unit):
        quantity = Units.Quantity(quantityStr)
        for key, setting in self.unit_system.items():
            unit = unit.replace(key, setting)
        return float(quantity.getValueAs(unit))

    def _handleConstants(self):
        self.constsdef = {
            "Gravity": constants.gravity(),
            "StefanBoltzmann": constants.stefan_boltzmann(),
            "PermeabilityOfVacuum": constants.vacuum_permeability(),
            "PermittivityOfVacuum": constants.vacuum_permittivity(),
            "BoltzmannConstant": constants.boltzmann_constant(),
        }

    def _handleRedifinedConstants(self):
        """
        redefine constants in self.constsdef according constant redefine objects
        """
        objs = self.getMember("Fem::ConstantVacuumPermittivity")
        if len(objs) == 1:
            permittivity = float(objs[0].VacuumPermittivity.getValueAs("F/m"))
            # since the base unit of FC is in mm, we must scale it to get plain SI
            permittivity = permittivity * 1e-9
            Console.PrintLog(f"Overwriting vacuum permittivity with: {permittivity}\n")
            self.constsdef["PermittivityOfVacuum"] = "{} {}".format(permittivity, "F/m")
            self.handled(objs[0])
        elif len(objs) > 1:
            Console.PrintError(
                "More than one permittivity constant overwriting objects ({} objs). "
                "The permittivity constant overwriting is ignored.\n".format(len(objs))
            )

    def _handleSimulation(self):
        self._simulation("Coordinate System", self.solver.CoordinateSystem)
        self._simulation("Coordinate Mapping", (1, 2, 3))
        # Elmer uses SI base units, but our mesh is in mm, therefore we must tell
        # the solver that we have another scale
        self._simulation("Coordinate Scaling", 0.001)
        self._simulation("Simulation Type", self.solver.SimulationType)
        if self.solver.SimulationType == "Steady State":
            self._simulation("Steady State Max Iterations", self.solver.SteadyStateMaxIterations)
            self._simulation("Steady State Min Iterations", self.solver.SteadyStateMinIterations)
        if self.solver.SimulationType == "Scanning" or self.solver.SimulationType == "Transient":
            self._simulation("BDF Order", self.solver.BDFOrder)
            self._simulation("Output Intervals", self.solver.OutputIntervals)
            self._simulation("Timestep Intervals", self.solver.TimestepIntervals)
            self._simulation("Timestep Sizes", self.solver.TimestepSizes)
            self._simulation("Timestepping Method", "BDF")
        self._simulation("Use Mesh Names", True)

    def _writeStartinfo(self):
        path = os.path.join(self.directory, _STARTINFO_NAME)
        with open(path, "w") as f:
            f.write(f"{_SIF_NAME}\n")

    # -------------------------------------------------------------------------------------------
    # Deformation

    def _handleDeformation(self):
        DEFW = DEF_writer.DeformationWriter(self, self.solver)
        activeIn = []
        for equation in self.solver.Group:
            if femutils.is_of_type(equation, "Fem::EquationElmerDeformation"):
                if not self._haveMaterialSolid():
                    raise WriteError(
                        "The Deformation equation requires at least one body with a solid material!"
                    )
                if equation.References:
                    activeIn = equation.References[0][1]
                else:
                    activeIn = self.getAllBodies()
                solverSection = DEFW.getDeformationSolver(equation)
                for body in activeIn:
                    if not self.isBodyMaterialFluid(body):
                        self._addSolver(body, solverSection)
                        DEFW.handleDeformationEquation(activeIn, equation)
        if activeIn:
            DEFW.handleDeformationConstants()
            DEFW.handleDeformationBndConditions()
            DEFW.handleDeformationInitial(activeIn)
            DEFW.handleDeformationBodyForces(activeIn)
            DEFW.handleDeformationMaterial(activeIn)

    # -------------------------------------------------------------------------------------------
    # Elasticity

    def _handleElasticity(self):
        ELW = EL_writer.ElasticityWriter(self, self.solver)
        activeIn = []
        for equation in self.solver.Group:
            if femutils.is_of_type(equation, "Fem::EquationElmerElasticity"):
                if not self._haveMaterialSolid():
                    raise WriteError(
                        "The Elasticity equation requires at least one body with a solid material!"
                    )
                if equation.References:
                    activeIn = equation.References[0][1]
                else:
                    activeIn = self.getAllBodies()
                solverSection = ELW.getElasticitySolver(equation)
                for body in activeIn:
                    if not self.isBodyMaterialFluid(body):
                        self._addSolver(body, solverSection)
                        ELW.handleElasticityEquation(activeIn, equation)
        if activeIn:
            ELW.handleElasticityConstants()
            ELW.handleElasticityBndConditions()
            ELW.handleElasticityInitial(activeIn)
            ELW.handleElasticityBodyForces(activeIn)
            ELW.handleElasticityMaterial(activeIn)

    # -------------------------------------------------------------------------------------------
    # Electrostatic

    def _handleElectrostatic(self):
        ESW = ES_writer.ESwriter(self, self.solver)
        activeIn = []
        for equation in self.solver.Group:
            if femutils.is_of_type(equation, "Fem::EquationElmerElectrostatic"):
                if equation.References:
                    activeIn = equation.References[0][1]
                else:
                    activeIn = self.getAllBodies()
                solverSection = ESW.getElectrostaticSolver(equation)
                for body in activeIn:
                    self._addSolver(body, solverSection)
        if activeIn:
            ESW.handleElectrostaticConstants()
            ESW.handleElectrostaticBndConditions()
            ESW.handleElectrostaticBodyForces()
            ESW.handleElectrostaticMaterial(activeIn)

    # -------------------------------------------------------------------------------------------
    # Electricforce

    def _handleElectricforce(self):
        EFW = EF_writer.EFwriter(self, self.solver)
        activeIn = []
        for equation in self.solver.Group:
            if femutils.is_of_type(equation, "Fem::EquationElmerElectricforce"):
                if equation.References:
                    activeIn = equation.References[0][1]
                else:
                    activeIn = self.getAllBodies()
                solverSection = EFW.getElectricforceSolver(equation)
                for body in activeIn:
                    self._addSolver(body, solverSection)

    # -------------------------------------------------------------------------------------------
    # Flow

    def _handleFlow(self):
        FlowW = flow_writer.Flowwriter(self, self.solver)
        activeIn = []
        for equation in self.solver.Group:
            if femutils.is_of_type(equation, "Fem::EquationElmerFlow"):
                if not self._haveMaterialFluid():
                    raise WriteError(
                        "The Flow equation requires at least one body with a fluid material!"
                    )
                if equation.References:
                    activeIn = equation.References[0][1]
                else:
                    activeIn = self.getAllBodies()
                solverSection = FlowW.getFlowSolver(equation)
                for body in activeIn:
                    if self.isBodyMaterialFluid(body):
                        self._addSolver(body, solverSection)
                        FlowW.handleFlowEquation(activeIn, equation)
        if activeIn:
            FlowW.handleFlowConstants()
            FlowW.handleFlowBndConditions()
            FlowW.handleFlowInitialPressure(activeIn)
            FlowW.handleFlowInitialVelocity(activeIn)
            FlowW.handleFlowMaterial(activeIn)

    # -------------------------------------------------------------------------------------------
    # Flux

    def _handleFlux(self):
        FluxW = flux_writer.Fluxwriter(self, self.solver)
        activeIn = []
        for equation in self.solver.Group:
            if femutils.is_of_type(equation, "Fem::EquationElmerFlux"):
                if equation.References:
                    activeIn = equation.References[0][1]
                else:
                    activeIn = self.getAllBodies()
                solverSection = FluxW.getFluxSolver(equation)
                for body in activeIn:
                    self._addSolver(body, solverSection)

    # -------------------------------------------------------------------------------------------
    # Heat

    def _handleHeat(self):
        HeatW = heat_writer.Heatwriter(self, self.solver)
        activeIn = []
        for equation in self.solver.Group:
            if femutils.is_of_type(equation, "Fem::EquationElmerHeat"):
                if equation.References:
                    activeIn = equation.References[0][1]
                else:
                    activeIn = self.getAllBodies()
                solverSection = HeatW.getHeatSolver(equation)
                for body in activeIn:
                    self._addSolver(body, solverSection)
                HeatW.handleHeatEquation(activeIn, equation)
        if activeIn:
            HeatW.handleHeatConstants()
            HeatW.handleHeatBndConditions()
            HeatW.handleHeatInitial(activeIn)
            HeatW.handleHeatBodyForces(activeIn)
            HeatW.handleHeatMaterial(activeIn)

    # -------------------------------------------------------------------------------------------
    # Magnetodynamic

    def _handleMagnetodynamic(self):
        MgDyn = MgDyn_writer.MgDynwriter(self, self.solver)
        activeIn = []
        for equation in self.solver.Group:
            if femutils.is_of_type(equation, "Fem::EquationElmerMagnetodynamic"):
                if equation.References:
                    activeIn = equation.References[0][1]
                else:
                    activeIn = self.getAllBodies()

                solverSection = MgDyn.getMagnetodynamicSolver(equation)
                solverPostSection = MgDyn.getMagnetodynamicSolverPost(equation)
                for body in activeIn:
                    self._addSolver(body, solverSection)
                    self._addSolver(body, solverPostSection)
        if activeIn:
            MgDyn.handleMagnetodynamicConstants()
            MgDyn.handleMagnetodynamicBndConditions(equation)
            MgDyn.handleMagnetodynamicBodyForces(activeIn, equation)
            MgDyn.handleMagnetodynamicMaterial(activeIn)

    # -------------------------------------------------------------------------------------------
    # Magnetodynamic2D

    def _handleMagnetodynamic2D(self):
        MgDyn2D = MgDyn2D_writer.MgDyn2Dwriter(self, self.solver)
        activeIn = []
        for equation in self.solver.Group:
            if femutils.is_of_type(equation, "Fem::EquationElmerMagnetodynamic2D"):
                if equation.References:
                    activeIn = equation.References[0][1]
                else:
                    activeIn = self.getAllBodies()
                # Magnetodynamic2D cannot handle all coordinate sysytems
                if self.solver.CoordinateSystem in _COORDS_NON_MAGNETO_2D:
                    raise WriteError(
                        "The coordinate setting '{}'\n is not "
                        "supported by the equation 'Magnetodynamic2D'.\n\n"
                        "Possible is:\n'Cartesian 2D' or 'Axi Symmetric'".format(
                            self.solver.CoordinateSystem
                        )
                    )

                solverSection = MgDyn2D.getMagnetodynamic2DSolver(equation)
                solverPostSection = MgDyn2D.getMagnetodynamic2DSolverPost(equation)
                for body in activeIn:
                    self._addSolver(body, solverSection)
                    self._addSolver(body, solverPostSection)
                    MgDyn2D.handleMagnetodynamic2DEquation(activeIn, equation)
        if activeIn:
            MgDyn2D.handleMagnetodynamic2DConstants()
            MgDyn2D.handleMagnetodynamic2DBndConditions(equation)
            MgDyn2D.handleMagnetodynamic2DBodyForces(activeIn, equation)
            MgDyn2D.handleMagnetodynamic2DMaterial(activeIn)

    # -------------------------------------------------------------------------------------------
    # StaticCurrent

    def _handleStaticCurrent(self):
        SCW = SC_writer.SCwriter(self, self.solver)
        activeIn = []
        for equation in self.solver.Group:
            if femutils.is_of_type(equation, "Fem::EquationElmerStaticCurrent"):
                if equation.References:
                    activeIn = equation.References[0][1]
                else:
                    activeIn = self.getAllBodies()
                solverSection = SCW.getStaticCurrentSolver(equation)
                for body in activeIn:
                    self._addSolver(body, solverSection)
                SCW.handleStaticCurrentBodyForces(activeIn, equation)

        if activeIn:
            SCW.handleStaticCurrentConstants()
            SCW.handleStaticCurrentBndConditions()
            SCW.handleStaticCurrentMaterial(activeIn)

    # -------------------------------------------------------------------------------------------
    # Solver handling

    def createEmptySolver(self):
        s = sifio.createSection(sifio.SOLVER)
        return s

    def _updateLinearSolver(self, equation):
        if self._hasExpression(equation) != equation.LinearTolerance:
            equation.setExpression("LinearTolerance", str(equation.LinearTolerance))
        if self._hasExpression(equation) != equation.SteadyStateTolerance:
            equation.setExpression("SteadyStateTolerance", str(equation.SteadyStateTolerance))
        if equation.BiCGstablDegree == 0:
            equation.BiCGstablDegree = 2
        if not hasattr(equation, "LinearSystemSolverDisabled"):
            equation.addProperty(
                "App::PropertyBool",
                "LinearSystemSolverDisabled",
                "Linear System",
                (
                    "Disable the linear system.\n"
                    "Only use for special cases\n"
                    "and consult the Elmer docs."
                ),
                locked=True,
            )
        if not hasattr(equation, "IdrsParameter"):
            equation.addProperty(
                "App::PropertyIntegerConstraint",
                "IdrsParameter",
                "Linear System",
                "Parameter for iterative method 'Idrs'",
                locked=True,
            )
            equation.IdrsParameter = (2, 1, 10, 1)

    def createLinearSolver(self, equation):
        # first check if we have to update
        self._updateLinearSolver(equation)
        # write the solver
        s = sifio.createSection(sifio.SOLVER)
        s.priority = equation.Priority
        s["Linear System Solver"] = equation.LinearSolverType
        if equation.LinearSystemSolverDisabled is True:
            s["Linear System Solver Disabled"] = equation.LinearSystemSolverDisabled
        if equation.LinearSolverType == "Direct":
            s["Linear System Direct Method"] = equation.LinearDirectMethod
        elif equation.LinearSolverType == "Iterative":
            s["Linear System Iterative Method"] = equation.LinearIterativeMethod
            if equation.LinearIterativeMethod == "BiCGStabl":
                s["BiCGstabl polynomial degree"] = equation.BiCGstablDegree
            if equation.LinearIterativeMethod == "Idrs":
                s["Idrs Parameter"] = equation.IdrsParameter
            s["Linear System Max Iterations"] = equation.LinearIterations
            s["Linear System Convergence Tolerance"] = equation.LinearTolerance
            s["Linear System Preconditioning"] = equation.LinearPreconditioning
        s["Steady State Convergence Tolerance"] = equation.SteadyStateTolerance
        s["Linear System Abort Not Converged"] = False
        s["Linear System Residual Output"] = 1
        s["Linear System Precondition Recompute"] = 1
        return s

    def _updateNonlinearSolver(self, equation):
        if self._hasExpression(equation) != equation.NonlinearTolerance:
            equation.setExpression("NonlinearTolerance", str(equation.NonlinearTolerance))
        if self._hasExpression(equation) != equation.NonlinearNewtonAfterTolerance:
            equation.setExpression(
                "NonlinearNewtonAfterTolerance", str(equation.NonlinearNewtonAfterTolerance)
            )

    def createNonlinearSolver(self, equation):
        # first check if we have to update
        self._updateNonlinearSolver(equation)
        # write the linear solver
        s = self.createLinearSolver(equation)
        # write the nonlinear solver
        s["Nonlinear System Max Iterations"] = equation.NonlinearIterations
        s["Nonlinear System Convergence Tolerance"] = equation.NonlinearTolerance
        s["Nonlinear System Relaxation Factor"] = equation.RelaxationFactor
        s["Nonlinear System Newton After Iterations"] = equation.NonlinearNewtonAfterIterations
        s["Nonlinear System Newton After Tolerance"] = equation.NonlinearNewtonAfterTolerance
        return s

    # -------------------------------------------------------------------------------------------
    # Helper functions

    def _haveMaterialSolid(self):
        for obj in self.getMember("App::MaterialObject"):
            m = obj.Material
            # fluid material always has KinematicViscosity defined
            if "KinematicViscosity" not in m:
                return True
        return False

    def _haveMaterialFluid(self):
        for obj in self.getMember("App::MaterialObject"):
            m = obj.Material
            # fluid material always has a viscosity defined
            if ("DynamicViscosity" in m) or ("KinematicViscosity" in m):
                return True
        return False

    def isBodyMaterialFluid(self, body):
        # we can have the case that a body has no assigned material
        # then assume it is a solid
        if self.getBodyMaterial(body) is not None:
            m = self.getBodyMaterial(body).Material
            return ("DynamicViscosity" in m) or ("KinematicViscosity" in m)
        return False

    def getBodyMaterial(self, name):
        for obj in self.getMember("App::MaterialObject"):
            # we can have e.g. the case there are 2 bodies and 2 materials
            # body 2 has material 2 as reference while material 1 has no reference
            # therefore we must not return a material when it is not referenced
            if obj.References and (name in obj.References[0][1]):
                return obj
        # 'name' was not in the reference of any material
        return None

    def getDensity(self, m):
        density = self.convert(m["Density"], "M/L^3")
        if self.getMeshDimension() == 2:
            density *= 1e3
        return density

    def _hasExpression(self, equation):
        for obj, exp in equation.ExpressionEngine:
            if obj == equation:
                return exp
        return None

    def getUniqueVarName(self, varName):
        postfix = 1
        if varName in self._usedVarNames:
            varName += "%2d" % postfix
        while varName in self._usedVarNames:
            postfix += 1
            varName = varName[:-2] + "%2d" % postfix
        self._usedVarNames.add(varName)
        return varName

    def getAllBodies(self):
        obj = self.getSingleMember("Fem::FemMeshObject")
        bodyCount = 0
        prefix = ""
        if obj.Shape.Shape.Solids:
            prefix = "Solid"
            bodyCount = len(obj.Shape.Shape.Solids)
        elif obj.Shape.Shape.Faces:
            prefix = "Face"
            bodyCount = len(obj.Shape.Shape.Faces)
        elif obj.Shape.Shape.Edges:
            prefix = "Edge"
            bodyCount = len(obj.Shape.Shape.Edges)
        return [prefix + str(i + 1) for i in range(bodyCount)]

    def getMeshDimension(self):
        obj = self.getSingleMember("Fem::FemMeshObject")
        if obj.Shape.Shape.Solids:
            return 3
        if obj.Shape.Shape.Faces:
            return 2
        if obj.Shape.Shape.Edges:
            return 1
        return None

    def _addOutputSolver(self):
        s = sifio.createSection(sifio.SOLVER)
        # Since FreeCAD meshes are in mm we let Elmer scale it
        # _handleSimulation(self).
        # To get it back in the original size we let Elmer scale it back
        s["Coordinate Scaling Revert"] = True
        s["Equation"] = "ResultOutput"
        if self.solver.SimulationType == "Scanning" or self.solver.SimulationType == "Transient":
            # we must execute the post solver every time we output a result
            # therefore we must use the same as self.solver.OutputIntervals
            s["Exec Intervals"] = self.solver.OutputIntervals
        else:
            s["Exec Solver"] = "After simulation"
        s["Procedure"] = sifio.FileAttr("ResultOutputSolve/ResultOutputSolver")
        s["Output File Name"] = sifio.FileAttr("FreeCAD")
        s["Vtu Format"] = True
        s["Binary Output"] = self.solver.BinaryOutput
        s["Save Geometry Ids"] = self.solver.SaveGeometryIndex
        s["Vtu Time Collection"] = True
        if self.unit_schema == Units.Scheme.MKS:
            s["Coordinate Scaling Revert"] = True
            Console.PrintMessage(
                "'Coordinate Scaling Revert = Logical True' was "
                "inserted into the solver input file.\n"
            )
        for name in self.getAllBodies():
            self._addSolver(name, s)

    def _writeSif(self):
        sifPath = os.path.join(self.directory, _SIF_NAME)
        with open(sifPath, "w") as fstream:
            sif = sifio.Sif(self._builder)
            sif.write(fstream)

    def handled(self, obj):
        self._handledObjects.add(obj)

    def _simulation(self, key, attr):
        self._builder.simulation(key, attr)

    def constant(self, key, attr):
        self._builder.constant(key, attr)

    def initial(self, body, key, attr):
        self._builder.initial(body, key, attr)

    def material(self, body, key, attr):
        self._builder.material(body, key, attr)

    def equation(self, body, key, attr):
        self._builder.equation(body, key, attr)

    def bodyForce(self, body, key, attr):
        self._builder.bodyForce(body, key, attr)

    def _addSolver(self, body, solverSection):
        self._builder.addSolver(body, solverSection)

    def boundary(self, boundary, key, attr):
        self._builder.boundary(boundary, key, attr)

    def _addSection(self, section):
        self._builder.addSection(section)

    def getMember(self, t):
        return membertools.get_member(self.analysis, t)

    def getSingleMember(self, t):
        return membertools.get_single_member(self.analysis, t)


class WriteError(Exception):
    pass


##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides various functions to work with faces."""
## @package faces
# \ingroup draftgeoutils
# \brief Provides various functions to work with faces.

import lazy_loader.lazy_loader as lz

import DraftVecUtils
from FreeCAD import Base
from draftgeoutils.geometry import are_coplanar

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")

## \addtogroup draftgeoutils
# @{


def concatenate(shape):
    """Turn several faces into one."""
    boundary_edges = getBoundary(shape)
    sorted_edges = Part.sortEdges(boundary_edges)

    try:
        wires = [Part.Wire(edges) for edges in sorted_edges]
        face = Part.makeFace(wires, "Part::FaceMakerBullseye")
    except Base.FreeCADError:
        print(
            "DraftGeomUtils: Fails to join faces into one. "
            + "The precision of the faces would be insufficient"
        )
        return shape
    else:
        if not wires[0].isClosed():
            return wires[0]
        else:
            return face


def getBoundary(shape):
    """Return the boundary edges of a group of faces."""
    if isinstance(shape, list):
        shape = Part.makeCompound(shape)

    # Make a lookup-table where we get the number of occurrences
    # to each edge in the fused face
    table = dict()
    for f in shape.Faces:
        for e in f.Edges:
            hash_code = e.hashCode()
            if hash_code in table:
                table[hash_code] = table[hash_code] + 1
            else:
                table[hash_code] = 1

    # Filter out the edges shared by more than one sub-face
    bound = list()
    for e in shape.Edges:
        if table[e.hashCode()] == 1:
            bound.append(e)
    return bound


def is_coplanar(faces, tol=-1):
    """Return True if all faces in the given list are coplanar.

    Parameters
    ----------
    faces: list
        List of faces to check coplanarity.
    tol: float, optional
        It defaults to `-1`, the tolerance of confusion, equal to 1e-7.
        Is the maximum deviation to be considered coplanar.

    Returns
    -------
    out: bool
        True if all face are coplanar. False in other case.
    """

    first_face = faces[0]
    for face in faces:
        if not are_coplanar(first_face, face, tol):
            return False

    return True


isCoplanar = is_coplanar


def bind(w1, w2, per_segment=False):
    """Bind 2 wires by their endpoints and returns a face / compound of faces.

    If per_segment is True and the wires have the same number of edges, the
    wires are processed per segment: a separate face is created for each pair
    of edges (one from w1 and one from w2), and the faces are then fused. This
    avoids problems with walls based on wires that selfintersect, or that have
    a loop that ends in a T-connection (f.e. a wire shaped like a number 6).
    """

    def create_face(w1, w2):

        try:
            w3 = Part.LineSegment(w1.Vertexes[0].Point, w2.Vertexes[0].Point).toShape()
            w4 = Part.LineSegment(w1.Vertexes[-1].Point, w2.Vertexes[-1].Point).toShape()
        except Part.OCCError:
            print("DraftGeomUtils: unable to bind wires")
            return None
        if w3.section(w4).Vertexes:
            print("DraftGeomUtils: Problem, a segment is self-intersecting, please check!")
        f = Part.Face(Part.Wire(w1.Edges + [w3] + w2.Edges + [w4]))
        return f

    if not w1 or not w2:
        print("DraftGeomUtils: unable to bind wires")
        return None

    if per_segment and len(w1.Edges) > 1 and len(w1.Edges) == len(w2.Edges):
        faces = []
        faces_list = []
        for edge1, edge2 in zip(w1.Edges, w2.Edges):
            # Find touching edges due to ArchWall Align in opposite
            # directions, and/or opposite edge orientations.
            #
            # w1 o-----o            w1 o-----o            w1 o-----o
            #          | w1                  |                     |
            # w2 +-----x-----o w1   w2 +-----+            w2 +-----+
            #       w2 |                  w2 | w1               w2 | w1
            #          +-----+ w2            o-----o w1            +-----+ w2
            #                                |                     |
            #                                +-----+ w2            o-----o w1
            #
            # TODO Maybe those edge pair should not be generated in offsetWire()
            #      and separate wires should then be returned.

            # If edges touch the Shape.section() compound will have 1 or 2 vertexes:
            if edge1.section(edge2).Vertexes:
                faces_list.append(faces)  # Break into separate list
                faces = []  # Reset original faces variable
                continue  # Skip the touching edge pair
            else:
                face = create_face(edge1, edge2)
                if face is None:
                    return None
                faces.append(face)

        # Usually there is last series of face after above 'for' routine,
        # EXCEPT when the last edge pair touch, faces had been appended
        # to faces_list, and reset faces =[]
        #
        # TODO Need fix further anything if there is a empty [] in faces_list ?
        #
        if faces_list and faces:
            # if wires are closed, 1st & last series of faces might be connected
            # except when
            # 1) there are only 2 series, connecting would return invalid shape
            # 2) 1st series of faces happens to be [], i.e. 1st edge pairs touch
            #
            if w1.isClosed() and w2.isClosed() and len(faces_list) > 1 and faces_list[0]:
                faces_list[0].extend(
                    faces
                )  # TODO: To be reviewed, 'afterthought' on 2025.3.29, seems by 'extend', faces in 1st and last faces are not in sequential order
            else:
                faces_list.append(faces)  # Break into separate list
        from collections import Counter

        if faces_list:
            faces_fused_list = []
            for faces in faces_list:
                dir = []
                countDir = None
                for f in faces:
                    dir.append(f.normalAt(0, 0).z)
                countDir = Counter(dir)
                l = len(faces)
                m = max(countDir.values())  # max(countDir, key=countDir.get)
                if m != l:
                    print(
                        "DraftGeomUtils: Problem, the direction of "
                        + str(l - m)
                        + " out of "
                        + str(l)
                        + " segment is reversed, please check!"
                    )
                if len(faces) > 1:
                    # Below not good if a face is self-intersecting or reversed
                    # faces_fused = faces[0].fuse(faces[1:]).removeSplitter().Faces[0]
                    rf = faces[0]
                    for f in faces[1:]:
                        rf = rf.fuse(f).removeSplitter().Faces[0]
                        # rf = rf.fuse(f)  # Not working
                    # rf = rf.removeSplitter().Faces[0]  # Not working
                    faces_fused_list.append(rf)
                # faces might be empty list [], see above; skip if empty
                elif faces:
                    faces_fused_list.append(faces[0])  # Only 1 face

            return Part.Compound(faces_fused_list)
        else:
            dir = []
            countDir = None
            for f in faces:
                dir.append(f.normalAt(0, 0).z)
            countDir = Counter(dir)
            l = len(faces)
            m = max(countDir.values())  # max(countDir, key=countDir.get)
            if m != l:
                print(
                    "DraftGeomUtils: Problem, the direction of "
                    + str(l - m)
                    + " out of "
                    + str(l)
                    + " segment is reversed, please check!"
                )
            # Below not good if a face is self-intersecting or reversed
            # return faces[0].fuse(faces[1:]).removeSplitter().Faces[0]
            rf = faces[0]
            for f in faces[1:]:
                rf = rf.fuse(f).removeSplitter().Faces[0]
                # rf = rf.fuse(f)  # Not working
            # rf = rf.removeSplitter().Faces[0]  # Not working
            return rf

    elif w1.isClosed() and w2.isClosed():
        d1 = w1.BoundBox.DiagonalLength
        d2 = w2.BoundBox.DiagonalLength
        if d1 < d2:
            w1, w2 = w2, w1
        # return Part.Face(w1).cut(Part.Face(w2)).Faces[0] # Only works if wires do not self-intersect.
        try:
            face = Part.Face([w1, w2])
            face.fix(1e-7, 0, 1)
            return face
        except Part.OCCError:
            print("DraftGeomUtils: unable to bind wires")
            return None
    else:
        return create_face(w1, w2)


def cleanFaces(shape):
    """Remove inner edges from coplanar faces."""
    faceset = shape.Faces

    def find(hc):
        """Find a face with the given hashcode."""
        for f in faceset:
            if f.hashCode() == hc:
                return f

    def findNeighbour(hface, hfacelist):
        """Find the first neighbour of a face, and return its index."""
        eset = []
        for e in find(hface).Edges:
            eset.append(e.hashCode())
        for i in range(len(hfacelist)):
            for ee in find(hfacelist[i]).Edges:
                if ee.hashCode() in eset:
                    return i
        return None

    # build lookup table
    lut = {}
    for face in faceset:
        for edge in face.Edges:
            if edge.hashCode() in lut:
                lut[edge.hashCode()].append(face.hashCode())
            else:
                lut[edge.hashCode()] = [face.hashCode()]

    # print("lut:",lut)
    # take edges shared by 2 faces
    sharedhedges = []
    for k, v in lut.items():
        if len(v) == 2:
            sharedhedges.append(k)

    # print(len(sharedhedges)," shared edges:",sharedhedges)
    # find those with same normals
    targethedges = []
    for hedge in sharedhedges:
        faces = lut[hedge]
        n1 = find(faces[0]).normalAt(0.5, 0.5)
        n2 = find(faces[1]).normalAt(0.5, 0.5)
        if n1 == n2:
            targethedges.append(hedge)

    # print(len(targethedges)," target edges:",targethedges)
    # get target faces
    hfaces = []
    for hedge in targethedges:
        for f in lut[hedge]:
            if f not in hfaces:
                hfaces.append(f)

    # print(len(hfaces)," target faces:",hfaces)
    # sort islands
    islands = [[hfaces.pop(0)]]
    currentisle = 0
    currentface = 0
    found = True
    while hfaces:
        if not found:
            if len(islands[currentisle]) > (currentface + 1):
                currentface += 1
                found = True
            else:
                islands.append([hfaces.pop(0)])
                currentisle += 1
                currentface = 0
                found = True
        else:
            f = findNeighbour(islands[currentisle][currentface], hfaces)
            if f is not None:
                islands[currentisle].append(hfaces.pop(f))
            else:
                found = False

    # print(len(islands)," islands:",islands)
    # make new faces from islands
    newfaces = []
    treated = []
    for isle in islands:
        treated.extend(isle)
        fset = []
        for i in isle:
            fset.append(find(i))
        bounds = getBoundary(fset)
        shp = Part.Wire(Part.__sortEdges__(bounds))
        shp = Part.Face(shp)
        if shp.normalAt(0.5, 0.5) != find(isle[0]).normalAt(0.5, 0.5):
            shp.reverse()
        newfaces.append(shp)

    # print("new faces:",newfaces)
    # add remaining faces
    for f in faceset:
        if not f.hashCode() in treated:
            newfaces.append(f)

    # print("final faces")
    # finishing
    fshape = Part.makeShell(newfaces)
    if shape.isClosed():
        fshape = Part.makeSolid(fshape)
    return fshape


def removeSplitter(shape):
    """Return a face from removing the splitter in a list of faces.

    This is an alternative, shared edge-based version of Part.removeSplitter.
    Returns a face, or `None` if the operation failed.
    """
    lookup = dict()
    for f in shape.Faces:
        for e in f.Edges:
            h = e.hashCode()
            if h in lookup:
                lookup[h].append(e)
            else:
                lookup[h] = [e]

    edges = [e[0] for e in lookup.values() if len(e) == 1]

    try:
        face = Part.Face(Part.Wire(edges))
    except Part.OCCError:
        # operation failed
        return None
    else:
        if face.isValid():
            return face

    return None


## @}
